"""Test insteonplm X10 devices."""
import asyncio
import logging

from insteonplm.constants import (X10_COMMAND_ON,
                                  X10_COMMAND_OFF,
                                  X10_COMMAND_DIM,
                                  X10_COMMAND_BRIGHT,
                                  X10_COMMAND_ALL_UNITS_OFF,
                                  X10_COMMAND_ALL_LIGHTS_ON,
                                  X10_COMMAND_ALL_LIGHTS_OFF)
from insteonplm.devices.x10 import (X10OnOff,
                                    X10Dimmable,
                                    X10AllUnitsOff,
                                    X10AllLightsOff,
                                    X10AllLightsOn)
from insteonplm.messages.x10send import X10Send
from insteonplm.messages.x10received import X10Received
import insteonplm.utils
from .mockCallbacks import MockCallbacks
from .mockPLM import MockPLM


_LOGGER = logging.getLogger()


def test_x10OnOff():
    @asyncio.coroutine
    def run_test(loop):
        housecode = 'C' # byte 0x02
        unitcode = 9    # byte 0x07
        plm = MockPLM(loop)
        device = X10OnOff(plm, housecode, unitcode)
        plm.devices[device.address.id] = device
        _LOGGER.debug('X10 device id: %s', device.address.id)
        assert device.address.human == 'X10.{}.{:02d}'.format(housecode, unitcode)

        # Send On command and test both commands sent
        device.states[0x01].on()
        yield from asyncio.sleep(.1, loop=loop)
        assert plm.sentmessage == '02632700'
        msg = X10Send(0x27, 0x00, 0x06)
        device.receive_message(msg)
        yield from  asyncio.sleep(.1, loop=loop)
        assert plm.sentmessage == '02632280'
        yield from  asyncio.sleep(.1, loop=loop)
        msg = X10Send(0x22, 0x00, 0x06)
        device.receive_message(msg)
        yield from  asyncio.sleep(.1, loop=loop)

        # Send Off command and test both commands sent
        device.states[0x01].off()
        yield from asyncio.sleep(.1, loop=loop)
        assert plm.sentmessage == '02632700'
        msg = X10Send(0x27, 0x00, 0x06)
        device.receive_message(msg)
        yield from  asyncio.sleep(.1, loop=loop)
        assert plm.sentmessage == '02632380'
        yield from  asyncio.sleep(.1, loop=loop)
        msg = X10Send(0x23, 0x00, 0x06)
        device.receive_message(msg)
        yield from  asyncio.sleep(.1, loop=loop)
    
    loop = asyncio.get_event_loop()
    loop.run_until_complete(run_test(loop))


def test_dimmable():
    @asyncio.coroutine
    def run_test(loop):
        housecode = 'C' # byte 0x02
        unitcode = 9    # byte 0x07
        newval = 200
        steps = round(newval / (255/22)) + 1
        _LOGGER.info('Steps %d', steps)
        plm = MockPLM(loop)
        cb = MockCallbacks()
        device = X10Dimmable(plm, housecode, unitcode)
        device.states[0x01].register_updates(cb.callbackmethod1)
        plm.devices[device.address.id] = device

        device.states[0x01].set_level(200)
        for i in range(0, steps):
            _LOGGER.info('Sending ACK messages')
            msg = X10Send(0x27, 0x00, 0x06)
            device.receive_message(msg)
            yield from  asyncio.sleep(.1, loop=loop)

            yield from  asyncio.sleep(.1, loop=loop)
            msg = X10Send(0x25, 0x80, 0x06)
        _LOGGER.info('New value: 0x%02x', cb.callbackvalue1)
        assert cb.callbackvalue1 == round((steps -1)*(255 / 22))
    
    loop = asyncio.get_event_loop()
    loop.run_until_complete(run_test(loop))


def test_on_received():
    @asyncio.coroutine
    def run_test(loop):
        housecode = 'C' # byte 0x02
        unitcode = 9    # byte 0x07
        plm = MockPLM(loop)
        cb = MockCallbacks()
        device = X10Dimmable(plm, housecode, unitcode, 22)
        plm.devices[device.address.id] = device
        device.states[0x01].register_updates(cb.callbackmethod1)

        msg = X10Received.command_msg(housecode, X10_COMMAND_ON)
        device.receive_message(msg)
        yield from asyncio.sleep(.1, loop=loop)
        assert cb.callbackvalue1 == 0xff

        msg = X10Received.command_msg(housecode, X10_COMMAND_OFF)
        device.receive_message(msg)
        yield from asyncio.sleep(.1, loop=loop)
        assert cb.callbackvalue1 == 0x00

    loop = asyncio.get_event_loop()
    loop.run_until_complete(run_test(loop))


def test_all_unit_types():
    plm = MockPLM()
    all_units_off = X10AllUnitsOff(plm, 'A', 20)
    all_lights_off = X10AllLightsOff(plm, 'A', 22)
    all_lights_on = X10AllLightsOn(plm, 'A', 21)

    assert all_units_off.description == 'X10 All Units Off Device'
    assert all_lights_off.description == 'X10 All Lights Off Device'
    assert all_lights_on.description == 'X10 All Lights On Device'
    assert all_units_off.id == 'x10A20'


def test_all_units_on_off():
    @asyncio.coroutine
    def run_test(loop):
        plm = MockPLM(loop)
        callbacks = MockCallbacks()
        all_units_off = X10AllUnitsOff(plm, 'A', 20)
        all_lights_off = X10AllLightsOff(plm, 'A', 22)
        all_lights_on = X10AllLightsOn(plm, 'A', 21)

        all_units_off.states[0x01].register_updates(callbacks.callbackmethod1)
        all_lights_off.states[0x01].register_updates(callbacks.callbackmethod2)
        all_lights_on.states[0x01].register_updates(callbacks.callbackmethod3)

        msg = X10Received.command_msg('A', X10_COMMAND_ALL_UNITS_OFF)
        plm.message_received(msg)
        _LOGGER.debug('Should have 1st callback')
        yield from asyncio.sleep(.1, loop=loop)
        assert callbacks.callbackvalue1 == 0x00
        _LOGGER.debug('Should have 2nd callback')
        yield from asyncio.sleep(2, loop=loop)
        assert callbacks.callbackvalue1 == 0xff

    loop = asyncio.get_event_loop()
    loop.run_until_complete(run_test(loop))


"""INSTEON General Controller Device Class."""
import asyncio
import logging

from insteonplm.constants import (X10_COMMAND_ALL_UNITS_OFF,
                                  X10_COMMAND_ALL_LIGHTS_ON,
                                  X10_COMMAND_ALL_LIGHTS_OFF)
from insteonplm.devices import X10Device
from insteonplm.messages.x10received import X10Received
from insteonplm.states.x10 import (X10OnOffSwitch, X10DimmableSwitch,
                                   X10OnOffSensor, X10AllUnitsOffSensor,
                                   X10AllLightsOnSensor, X10AllLightsOffSensor)


_LOGGER = logging.getLogger(__name__)


class X10OnOff(X10Device):
    """General X10 On / Off Switch Device Class."""

    def __init__(self, plm, housecode, unitcode):
        super().__init__(plm, housecode, unitcode)
        self._description = 'X10 On / Off Device'

        self._stateList[0x01] = X10OnOffSwitch(
            self._address, "x10OnOffSwitch", 0x01, self._send_msg,
            self._message_callbacks, 0x00)


class X10Dimmable(X10Device):
    """General X10 Dimmable Switch Device Class."""

    def __init__(self, plm, housecode, unitcode, dim_steps=22):
        super().__init__(plm, housecode, unitcode)
        self._description = 'X10 Dimmable Device'

        self._stateList[0x01] = X10DimmableSwitch(
            self._address, "x10DimmableSwitch", 0x01, self._send_msg,
            self._message_callbacks, 0x00)


class X10Sensor(X10Device):
    """General X10 On / Off Sensor Device Class."""

    def __init__(self, plm, housecode, unitcode, dim_steps=22):
        super().__init__(plm, housecode, unitcode)
        self._description = 'X10 On / Off Sensor Device'

        self._stateList[0x01] = X10OnOffSensor(
            self._address, "x10OnOffSensor", 0x01, self._send_msg,
            self._message_callbacks, 0x00)


class X10AllUnitsOff(X10Device):
    """X10 All Units Off Device."""

    def __init__(self, plm, housecode, unitcode):
        super().__init__(plm, housecode, 20)
        self._description = 'X10 All Units Off Device'

        self._stateList[0x01] = X10AllUnitsOffSensor(
            self._address, "x10AllUnitsOffSensor", 0x01, self._send_msg,
            self._message_callbacks, 0xff)
        
        self._stateList[0x01].register_updates(self._reset_state)
        self._register_messages()

    def _reset_state(self, addr, name, val):
        loop = self._plm.loop
        if val == 0x00:
            asyncio.ensure_future(self._reset_state_value(loop), loop=loop)

    @asyncio.coroutine
    def _reset_state_value(self, loop):
        yield from asyncio.sleep(1, loop=loop)
        self._stateList[0x01].reset()

    def _register_messages(self):
        all_units_off_msg = X10Received.command_msg(self.address.x10_housecode,
                                                    X10_COMMAND_ALL_UNITS_OFF)
        self._plm.message_callbacks.add(all_units_off_msg,
                                        self.receive_message)


class X10AllLightsOn(X10Device):
    """X10 All Lights On Device."""

    def __init__(self, plm, housecode, unitcode):
        super().__init__(plm, housecode, 21)
        self._description = 'X10 All Lights On Device'

        self._stateList[0x01] = X10AllLightsOnSensor(
            self._address, "X10AllLightsOnSensor", 0x01, self._send_msg,
            self._message_callbacks, 0x00)

        self._stateList[0x01].register_updates(self._reset_state)
        self._register_messages()

    def _reset_state(self, addr, name, val):
        loop = self._plm.loop
        if val == 0xff:
            asyncio.ensure_future(self._reset_state_value(loop), loop=loop)

    @asyncio.coroutine
    def _reset_state_value(self, loop):
        yield from asyncio.sleep(1, loop=loop)
        self._stateList[0x01].reset()

    def _register_messages(self):
        all_on_msg = X10Received.command_msg(self.address.x10_housecode,
                                             X10_COMMAND_ALL_LIGHTS_ON)
        self._plm.message_callbacks.add(all_on_msg,
                                        self.receive_message)


class X10AllLightsOff(X10Device):
    """X10 All Lights Off Device."""

    def __init__(self, plm, housecode, unitcode):
        super().__init__(plm, housecode, 22)
        self._description = 'X10 All Lights Off Device'

        self._stateList[0x01] = X10AllLightsOffSensor(
            self._address, "X10AllLightsOffSensor", 0x01, self._send_msg,
            self._message_callbacks, 0xff)

        self._stateList[0x01].register_updates(self._reset_state)
        self._register_messages()

    def _reset_state(self, addr, name, val):
        loop = self._plm.loop
        if val == 0x00:
            asyncio.ensure_future(self._reset_state_value(loop), loop=loop)

    @asyncio.coroutine
    def _reset_state_value(self, loop):
        yield from asyncio.sleep(1, loop=loop)
        self._stateList[0x01].reset()

    def _register_messages(self):
        all_off_msg = X10Received.command_msg(self.address.x10_housecode,
                                              X10_COMMAND_ALL_LIGHTS_OFF)
        self._plm.message_callbacks.add(all_off_msg,
                                        self.receive_message)



"""Test Security Health and Saftey devices."""
import asyncio

from insteonplm.constants import (COMMAND_LIGHT_OFF_0X13_0X00,
                                  COMMAND_LIGHT_ON_0X11_NONE,
                                  MESSAGE_FLAG_BROADCAST_0X80,
                                  MESSAGE_TYPE_ALL_LINK_CLEANUP,
                                  MESSAGE_TYPE_ALL_LINK_BROADCAST)
from insteonplm.devices.securityHealthSafety import (
    SecurityHealthSafety, SecurityHealthSafety_2842_222,
    SecurityHealthSafety_2845_222, SecurityHealthSafety_2852_222,
    SecurityHealthSafety_2982_222)
from insteonplm.messages.standardReceive import StandardReceive
from insteonplm.messages.messageFlags import MessageFlags
from .mockPLM import MockPLM
from .mockCallbacks import MockCallbacks

@asyncio.coroutine
def _onOffSenorTest(onOffClass, loop):
    """Test on/off sensor."""
    plm = MockPLM(loop=loop)
    address = '1a2b3c'
    target = bytearray([0x00, 0x00, 0x01])

    cat = 0x10
    subcat = 0x00
    product_key = 0x00
    description = 'Generic Security, Heath and Safety Device'
    model = ''

    callbacks = MockCallbacks()

    device = onOffClass(plm, address, cat, subcat, product_key,
                        description, model)
    plm.devices[address] = device
    device.states[0x01].register_updates(callbacks.callbackmethod1)
    msg = StandardReceive(
        address, target, COMMAND_LIGHT_ON_0X11_NONE, cmd2=0x01,
        flags=MessageFlags.create(MESSAGE_TYPE_ALL_LINK_BROADCAST, 0, 3, 3))
    plm.message_received(msg)
    yield from asyncio.sleep(.1, loop=loop)
    assert callbacks.callbackvalue1 == 1

    device = onOffClass(plm, address, cat, subcat, product_key,
                        description, model)
    device.states[0x01].register_updates(callbacks.callbackmethod1)
    msg = StandardReceive(
        address, target, COMMAND_LIGHT_OFF_0X13_0X00,
        flags=MessageFlags.create(MESSAGE_TYPE_ALL_LINK_BROADCAST, 0, 3, 3))
    plm.message_received(msg)
    yield from asyncio.sleep(.1, loop=loop)
    assert callbacks.callbackvalue1 == 0


def test_securityhealthsafety():
    """Test generic Securityhealthsafety devices."""
    @asyncio.coroutine
    def run_test(loop):
        """Asyncio coroutine to actually run the test."""
        class sensorState(object):
            """Callback class to caputure sensor value changes."""

            sensor = None

            def sensor_status_callback(self, device_id, state, value):
                """Callback method to update sensor value."""
                self.sensor = value

        plm = MockPLM(loop)
        address = '1a2b3c'
        target = bytearray([0x00, 0x00, 0x01])
        cmd2 = 0x01

        cat = 0x10
        subcat = 0x00
        product_key = 0x00
        description = 'Generic Security, Heath and Safety Device'
        model = ''

        callbacks = MockCallbacks()

        device = SecurityHealthSafety(plm, address, cat, subcat,
                                      product_key, description, model)
        plm.devices[address] = device
        device.states[0x01].register_updates(callbacks.callbackmethod1)
        msg = StandardReceive(
            address, target, COMMAND_LIGHT_ON_0X11_NONE, cmd2=cmd2,
            flags=MessageFlags.create(MESSAGE_TYPE_ALL_LINK_BROADCAST, 0, 3, 3))
        plm.message_received(msg)
        yield from asyncio.sleep(.1, loop=loop)
        assert callbacks.callbackvalue1 == cmd2

        device = SecurityHealthSafety(plm, address, cat, subcat,
                                      product_key, description, model)
        device.states[0x01].register_updates(callbacks.callbackmethod1)
        msg = StandardReceive(
            address, target, COMMAND_LIGHT_OFF_0X13_0X00,
            flags=MessageFlags.create(MESSAGE_TYPE_ALL_LINK_BROADCAST, 0, 3, 3))
        plm.message_received(msg)
        yield from asyncio.sleep(.1, loop=loop)
        assert callbacks.callbackvalue1 == 0x00

    loop = asyncio.get_event_loop()
    loop.run_until_complete(run_test(loop))



def test_securityhealthsafety_2982_222():
    """Test device 2982-222."""
    @asyncio.coroutine
    def run_test(loop):
        plm = MockPLM(loop)
        address = '1a2b3c'
        target = '4d5e6f'
        cmd2 = 0x04

        cat = 0x10
        subcat = 0x00
        product_key = 0x00
        description = 'Generic Security, Heath and Safety Device'
        model = ''

        callbacks = MockCallbacks()

        device = SecurityHealthSafety_2982_222(plm, address, cat, subcat,
                                               product_key, description, model)
        plm.devices[address] = device
        device.states[0x01].register_updates(callbacks.callbackmethod1)
        msg = StandardReceive(
            address, target, COMMAND_LIGHT_ON_0X11_NONE, cmd2=cmd2,
            flags=MessageFlags.create(MESSAGE_FLAG_BROADCAST_0X80, 0, 0, 0))
        plm.message_received(msg)
        yield from asyncio.sleep(.1, loop=loop)
        assert callbacks.callbackvalue1 == 0x6f

    loop = asyncio.get_event_loop()
    loop.run_until_complete(run_test(loop))


def test_securityHealthSafety_2842_222():
    """Test 2842-222."""
    loop = asyncio.get_event_loop()
    loop.run_until_complete(
        _onOffSenorTest(SecurityHealthSafety_2842_222, loop))


def test_securityHealthSafety_2845_2222():
    """Test device 2845-222."""
    loop = asyncio.get_event_loop()
    loop.run_until_complete(
        _onOffSenorTest(SecurityHealthSafety_2845_222, loop))


def test_securityHealthSafety_2852_222():
    """Test device 2852-222."""
    def _run_test(loop):
        """Test on/off sensor."""
        plm = MockPLM(loop)
        address = '1a2b3c'
        target = '4d5e6f'
        cmd2 = 0x04

        cat = 0x10
        subcat = 0x00
        product_key = 0x00
        description = 'Generic Security, Heath and Safety Device'
        model = ''

        callbacks = MockCallbacks()

        device = SecurityHealthSafety_2852_222(plm, address, cat, subcat,
                                               product_key, description, model)
        plm.devices[address] = device
        device.states[0x01].register_updates(callbacks.callbackmethod1)
        device.states[0x02].register_updates(callbacks.callbackmethod2)
        device.states[0x04].register_updates(callbacks.callbackmethod4)

        # Test Dry message received
        msg = StandardReceive(
            address=address, 
            target=bytearray([0x00, 0x00, 0x01]), 
            commandtuple=COMMAND_LIGHT_ON_0X11_NONE, cmd2=0x01,
            flags=MessageFlags.create(MESSAGE_TYPE_ALL_LINK_BROADCAST, 0, 3, 3))
        plm.message_received(msg)
        yield from asyncio.sleep(.1, loop=loop)
        assert callbacks.callbackvalue1 == 1
        assert callbacks.callbackvalue2 == 0
        assert callbacks.callbackvalue4 == 0x11

        # Test wet message received
        msg = StandardReceive(
            address=address, 
            target=bytearray([0x00, 0x00, 0x02]), 
            commandtuple=COMMAND_LIGHT_ON_0X11_NONE, cmd2=0x02,
            flags=MessageFlags.create(MESSAGE_TYPE_ALL_LINK_BROADCAST, 0, 3, 3))
        plm.message_received(msg)
        yield from asyncio.sleep(.1, loop=loop)
        assert callbacks.callbackvalue1 == 0
        assert callbacks.callbackvalue2 == 1
        assert callbacks.callbackvalue4 == 0x13
        
        # Test dry heartbeat received
        msg = StandardReceive(
            address=address, 
            target=bytearray([0x00, 0x00, 0x04]), 
            commandtuple=COMMAND_LIGHT_ON_0X11_NONE, cmd2=0x04,
            flags=MessageFlags.create(MESSAGE_TYPE_ALL_LINK_BROADCAST, 0, 3, 3))
        plm.message_received(msg)
        yield from asyncio.sleep(.1, loop=loop)
        assert callbacks.callbackvalue1 == 1
        assert callbacks.callbackvalue2 == 0
        
        # Test wet heartbeat received
        msg = StandardReceive(
            address=address, 
            target=bytearray([0x00, 0x00, 0x04]), 
            commandtuple={'cmd1': 0x13, 'cmd2': 0x04},
            flags=MessageFlags.create(MESSAGE_TYPE_ALL_LINK_BROADCAST, 0, 3, 3))
        plm.message_received(msg)
        yield from asyncio.sleep(.1, loop=loop)
        assert callbacks.callbackvalue1 == 0
        assert callbacks.callbackvalue2 == 1
        assert callbacks.callbackvalue4 == 0x13

    loop = asyncio.get_event_loop()
    loop.run_until_complete(
        _run_test(loop))


"""Insteon Device Classes."""
import asyncio
import async_timeout
import collections
import datetime
from enum import Enum
import logging

from insteonplm.address import Address
from insteonplm.constants import (
    COMMAND_ID_REQUEST_0X10_0X00,
    COMMAND_PRODUCT_DATA_REQUEST_0X03_0X00,
    COMMAND_ASSIGN_TO_ALL_LINK_GROUP_0X01_NONE,
    COMMAND_DELETE_FROM_ALL_LINK_GROUP_0X02_NONE,
    COMMAND_FX_USERNAME_0X03_0X01,
    COMMAND_ENTER_LINKING_MODE_0X09_NONE,
    COMMAND_ENTER_UNLINKING_MODE_0X0A_NONE,
    COMMAND_GET_INSTEON_ENGINE_VERSION_0X0D_0X00,
    COMMAND_PING_0X0F_0X00,
    COMMAND_EXTENDED_READ_WRITE_ALDB_0X2F_0X00,
    MESSAGE_TYPE_DIRECT_MESSAGE,
    MESSAGE_FLAG_DIRECT_MESSAGE_NAK_0XA0
)
from insteonplm.messagecallback import MessageCallback
from insteonplm.messages.extendedReceive import ExtendedReceive
from insteonplm.messages.standardReceive import StandardReceive
from insteonplm.messages.extendedSend import ExtendedSend
from insteonplm.messages.messageFlags import MessageFlags
from insteonplm.messages.standardSend import StandardSend
from insteonplm.messages.userdata import Userdata
from insteonplm.states import State

DIRECT_ACK_WAIT_TIMEOUT = 3
ALDB_RECORD_TIMEOUT = 10
ALDB_RECORD_RETRIES = 20
ALDB_ALL_RECORD_TIMEOUT = 30
ALDB_ALL_RECORD_RETRIES = 5


def create(plm, address, cat, subcat, firmware=None):
    """Create a device from device info data."""
    from insteonplm.devices.ipdb import IPDB
    ipdb = IPDB()
    product = ipdb[[cat, subcat]]
    deviceclass = product.deviceclass
    device = None
    if deviceclass is not None:
        device = deviceclass(plm, address, cat, subcat,
                             product.product_key,
                             product.description,
                             product.model)
    return device


def create_x10(plm, housecode, unitcode, feature):
    """Create an X10 device from a feature definition."""
    from insteonplm.devices.ipdb import IPDB
    ipdb = IPDB()
    product = ipdb.x10(feature)
    deviceclass = product.deviceclass
    device = None
    if deviceclass:
        device = deviceclass(plm, housecode, unitcode)
    return device


class Device(object):
    """INSTEON Device Class."""

    def __init__(self, plm, address, cat, subcat, product_key=0x00,
                 description='', model=''):
        """Initialize the Device class."""
        self.log = logging.getLogger(__name__)

        self._plm = plm

        self._address = Address(address)
        self._cat = cat
        self._subcat = subcat
        if self._subcat is None:
            self._subcat = 0x00
        self._product_key = product_key
        if self._product_key is None:
            self._product_key = 0x00
        self._description = description
        self._model = model

        self._last_communication_received = datetime.datetime(1, 1, 1, 1, 1, 1)
        self._product_data_in_aldb = False
        self._stateList = StateList()
        self._send_msg_lock = asyncio.Lock(loop=self._plm.loop)
        self._sent_msg_wait_for_directACK = {}
        self._directACK_received_queue = asyncio.Queue(loop=self._plm.loop)
        self._message_callbacks = MessageCallback()
        self._aldb = ALDB(self._send_msg, self._plm.loop, self._address)

        self._register_messages()

    # Public properties
    @property
    def address(self):
        """Return the INSTEON device address."""
        return self._address

    @property
    def cat(self):
        """Return the INSTEON device category."""
        return self._cat

    @property
    def subcat(self):
        """Return the INSTEON device subcategory."""
        return self._subcat

    @property
    def product_key(self):
        """Return the INSTEON product key."""
        return self._product_key

    @property
    def description(self):
        """Return the INSTEON device description."""
        return self._description

    @property
    def model(self):
        """Return the INSTEON device model number."""
        return self._model

    @property
    def id(self):
        """Return the ID of the device."""
        return self._address.id

    @property
    def states(self):
        """Return the device states/groups."""
        return self._stateList

    @property
    def prod_data_in_aldb(self):
        """Return if the PLM use the ALDB data to setup the device.

        True if Product data (cat, subcat) is stored in the PLM ALDB.
        False if product data must be aquired via a Device ID message or from a
        Product Data Request command.

        The method of linking determines if product data in the ALDB,
        therefore False is the default. The common reason to store product data
        in the ALDB is for one way devices or battery opperated devices where
        the ability to send a command request is limited.
        """
        return self._product_data_in_aldb

    @property
    def aldb(self):
        return self._aldb

    # Public Methods
    def async_refresh_state(self):
        """Request each state to provide status update."""
        for state in self._stateList:
            self._stateList[state].async_refresh_state()

    def id_request(self):
        """Request a device ID from a device."""
        msg = StandardSend(self.address, COMMAND_ID_REQUEST_0X10_0X00)
        self._plm.send_msg(msg)

    def product_data_request(self):
        """Request product data from a device.

        Not supported by all devices.
        Required after 01-Feb-2007.
        """
        msg = StandardSend(self._address,
                           COMMAND_PRODUCT_DATA_REQUEST_0X03_0X00)
        self._send_msg(msg)

    def assign_to_all_link_group(self, group=0x01):
        """Assign a device to an All-Link Group.

        The default is group 0x01.
        """
        msg = StandardSend(self._address,
                           COMMAND_ASSIGN_TO_ALL_LINK_GROUP_0X01_NONE,
                           cmd2=group)
        self._send_msg(msg)

    def delete_from_all_link_group(self, group):
        """Delete a device to an All-Link Group."""
        msg = StandardSend(self._address,
                           COMMAND_DELETE_FROM_ALL_LINK_GROUP_0X02_NONE,
                           cmd2=group)
        self._send_msg(msg)

    def fx_username(self):
        """Get FX Username.

        Only required for devices that support FX Commands.
        FX Addressee responds with an ED 0x0301 FX Username Response message
        """
        msg = StandardSend(self._address, COMMAND_FX_USERNAME_0X03_0X01)
        self._send_msg(msg)

    def device_text_string_request(self):
        """Get FX Username.

        Only required for devices that support FX Commands.
        FX Addressee responds with an ED 0x0301 FX Username Response message.
        """
        msg = StandardSend(self._address, COMMAND_FX_USERNAME_0X03_0X01)
        self._send_msg(msg)

    def enter_linking_mode(self, group=0x01):
        """Tell a device to enter All-Linking Mode.

        Same as holding down the Set button for 10 sec.
        Default group is 0x01.

        Not supported by i1 devices.
        """
        msg = StandardSend(self._address,
                           COMMAND_ENTER_LINKING_MODE_0X09_NONE,
                           cmd2=group)
        self._send_msg(msg)

    def enter_unlinking_mode(self, group):
        """Unlink a device from an All-Link group.

        Not supported by i1 devices.
        """
        msg = StandardSend(self._address,
                           COMMAND_ENTER_UNLINKING_MODE_0X0A_NONE,
                           cmd2=group)
        self._send_msg(msg)

    def get_engine_version(self):
        """Get the device engine version."""
        msg = StandardSend(self._address,
                           COMMAND_GET_INSTEON_ENGINE_VERSION_0X0D_0X00)
        self._send_msg(msg)

    def ping(self):
        """Ping a device."""
        msg = StandardSend(self._address, COMMAND_PING_0X0F_0X00)
        self._send_msg(msg)

    def read_aldb(self, mem_addr=0x0000, num_recs=0):
        """Read the device All-Link Database."""
        if self._aldb.version == ALDBVersion.Null:
            self.log.info('Device does not contain an ALDB')
        else:
            self.log.info('Reading ALDB for device %s', self._address)
            asyncio.ensure_future(self._aldb.load(mem_addr, num_recs),
                                  loop=self._plm.loop)
            self._aldb.add_loaded_callback(self._aldb_loaded_callback)

    def write_aldb(self, mem_addr: int, mode: str, group: int, target,
                   data1=0x00, data2=0x00, data3=0x00):
        """Write to the device All-Link Database.

        Paramters:
            Required:
            mode:   r - device is a responder of target
                    c - device is a controller of target
            group:  Link group
            target: Address of the other device

            Optional:
            data1:  Device dependant
            data2:  Device dependant
            data3:  Device dependant
        """
        if isinstance(mode, str) and mode.lower() in ['c', 'r']:
            pass
        else:
            self.log.info('mode: %s', mode)
            raise ValueError("Mode must be 'c' or 'r'")
        if isinstance(group, int):
            pass
        else:
            raise ValueError("Group must be an integer")

        target_addr = Address(target)

        self.log.info('calling aldb write_record')
        self._aldb.write_record(mem_addr, mode, group, target_addr,
                                data1, data2, data3)
        self._aldb.add_loaded_callback(self._aldb_loaded_callback)

    def del_aldb(self, mem_addr: int):
        """Delete an All-Link Database record."""
        self._aldb.del_record(mem_addr)
        self._aldb.add_loaded_callback(self._aldb_loaded_callback)

    def _handle_aldb_record_received(self, msg):
        self._aldb.record_received(msg)

    def _handle_pre_nak(self, msg):
        self.async_refresh_state()

    def _register_messages(self):
            ext_msg_aldb_record = ExtendedReceive.template(
                address=self._address,
                commandtuple=COMMAND_EXTENDED_READ_WRITE_ALDB_0X2F_0X00,
                userdata=Userdata.template({'d2': 1}),
                flags=MessageFlags.template(
                    messageType=MESSAGE_TYPE_DIRECT_MESSAGE,
                    extended=1))
            std_msg_pre_nak = StandardReceive.template(
                flags=MessageFlags.template(
                    messageType=MESSAGE_FLAG_DIRECT_MESSAGE_NAK_0XA0),
                cmd2=0xfc)
            ext_msg_pre_nak = ExtendedReceive.template(
                flags=MessageFlags.template(
                    messageType=MESSAGE_FLAG_DIRECT_MESSAGE_NAK_0XA0),
                cmd2=0xfc)
            self._message_callbacks.add(ext_msg_aldb_record,
                                        self._handle_aldb_record_received)
            self._message_callbacks.add(std_msg_pre_nak,
                                        self._handle_pre_nak)
            self._message_callbacks.add(ext_msg_pre_nak,
                                        self._handle_pre_nak)

    # Send / Receive message processing
    def receive_message(self, msg):
        self.log.debug('Starting Device.receive_message')
        if hasattr(msg, 'isack') and msg.isack:
            self.log.debug('Got Message ACK')
            if self._sent_msg_wait_for_directACK.get('callback') is not None:
                self.log.debug('Look for direct ACK')
                coro = self._wait_for_direct_ACK()
                asyncio.ensure_future(coro, loop=self._plm.loop)
            else:
                self.log.debug('DA queue: %s',
                               self._sent_msg_wait_for_directACK)
                self.log.debug('Message ACK with no callback')
        if (hasattr(msg, 'flags') and
                hasattr(msg.flags, 'isDirectACK') and
                msg.flags.isDirectACK):
            self.log.debug('Got Direct ACK message')
            if self._send_msg_lock.locked():
                self._directACK_received_queue.put_nowait(msg)
            else:
                self.log.debug('But Direct ACK not expected')
        callbacks = self._message_callbacks.get_callbacks_from_message(msg)
        for callback in callbacks:
            self.log.debug('Scheduling msg callback: %s', callback)
            self._plm.loop.call_soon(callback, msg)
        self._last_communication_received = datetime.datetime.now()
        self.log.debug('Ending Device.receive_message')

    def _send_msg(self, msg, callback=None, on_timeout=False):
        self.log.debug('Starting Device._send_msg')
        write_message_coroutine = self._process_send_queue(msg,
                                                           callback,
                                                           on_timeout)
        asyncio.ensure_future(write_message_coroutine,
                              loop=self._plm.loop)
        self.log.debug('Ending Device._send_msg')

    @asyncio.coroutine
    def _process_send_queue(self, msg, callback=None, on_timeout=False):
        self.log.debug('Starting Device._process_send_queue')
        yield from self._send_msg_lock
        if self._send_msg_lock.locked():
            self.log.debug("Lock is locked from yeild from")
        if callback:
            self._sent_msg_wait_for_directACK = {'msg': msg,
                                                 'callback': callback,
                                                 'on_timeout': on_timeout}
        self._plm.send_msg(msg)
        self.log.debug('Ending Device._process_send_queue')

    @asyncio.coroutine
    def _wait_for_direct_ACK(self):
        self.log.debug('Starting Device._wait_for_direct_ACK')
        msg = None
        while True:
            # wait for an item from the queue
            try:
                with async_timeout.timeout(DIRECT_ACK_WAIT_TIMEOUT):
                    msg = yield from self._directACK_received_queue.get()
                    break
            except asyncio.TimeoutError:
                self.log.debug('No direct ACK messages received.')
                break
        self.log.debug('Releasing lock')
        self._send_msg_lock.release()
        if msg or self._sent_msg_wait_for_directACK.get('on_timeout'):
            callback = self._sent_msg_wait_for_directACK.get('callback', None)
            if callback is not None:
                callback(msg)
        self._sent_msg_wait_for_directACK = {}
        self.log.debug('Ending Device._wait_for_direct_ACK')

    def _aldb_loaded_callback(self):
        self._plm.devices.save_device_info()


class X10Device(object):
    """X10 device class."""

    def __init__(self, plm, housecode, unitcode):
        """Initialize the X10Device class."""
        self._address = Address.x10(housecode, unitcode)
        self._plm = plm
        self._description = "Generic X10 device"
        self._model = ''
        self._aldb = ALDB(None, None, self._address, version=ALDBVersion.Null)
        self._message_callbacks = MessageCallback()
        self._stateList = StateList()
        self._send_msg_lock = asyncio.Lock(loop=self._plm.loop)
        self.log = logging.getLogger()

    @property
    def address(self):
        """X10 device address."""
        return self._address

    @property
    def description(self):
        """Return the INSTEON device description."""
        return self._description

    @property
    def model(self):
        """Return the INSTEON device model number."""
        return self._model

    @property
    def id(self):
        """Return the ID of the device."""
        return self._address.id

    @property
    def states(self):
        """Return the device states/groups."""
        return self._stateList

    @property
    def aldb(self):
        """Return the device All-Link Database."""
        return self._aldb

    # Send / Receive message processing
    def receive_message(self, msg):
        self.log.debug('Starting X10Device.receive_message')
        if hasattr(msg, 'isack') and msg.isack:
            self.log.debug('Got Message ACK')
            if self._send_msg_lock.locked():
                self._send_msg_lock.release()
        callbacks = self._message_callbacks.get_callbacks_from_message(msg)
        self.log.debug('Found %d callbacks for msg %s', len(callbacks), msg)
        for callback in callbacks:
            self.log.debug('Scheduling msg callback: %s', callback)
            self._plm.loop.call_soon(callback, msg)
        self._last_communication_received = datetime.datetime.now()
        self.log.debug('Ending Device.receive_message')

    def _send_msg(self, msg, wait_ack=True):
        self.log.debug('Starting Device._send_msg')
        write_message_coroutine = self._process_send_queue(msg, wait_ack)
        asyncio.ensure_future(write_message_coroutine,
                              loop=self._plm.loop)
        self.log.debug('Ending Device._send_msg')

    @asyncio.coroutine
    def _process_send_queue(self, msg, wait_ack):
        self.log.debug('Starting Device._process_send_queue')
        yield from self._send_msg_lock
        if self._send_msg_lock.locked():
            self.log.debug("Lock is locked from yeild from")

        self._plm.send_msg(msg, wait_timeout=2)
        if not wait_ack:
            self._send_msg_lock.release()
        self.log.debug('Ending Device._process_send_queue')


class StateList(object):
    """Internal class used to hold a list of device states."""

    def __init__(self):
        """Initialize the StateList Class."""
        self._stateList = {}

    def __len__(self):
        """Get the number of states in the StateList."""
        return len(self._stateList)

    def __iter__(self):
        """Iterate through each state in the StateList."""
        for state in self._stateList:
            yield state

    def __getitem__(self, group):
        """Get a state from the StateList."""
        return self._stateList.get(group, None)

    def __setitem__(self, group, state):
        """Add or update a state in the StateList."""
        if not isinstance(state, State):
            return ValueError

        self._stateList[group] = state

    def __repr__(self):
        """Juman representation of a state in the StateList."""
        attrs = vars(self)
        return ', '.join("%s: %r" % item for item in attrs.items())

    def add(self, plm, device, stateType, stateName, group, defaultValue=None):
        """Add a state to the StateList."""
        self._stateList[group] = stateType(plm, device, stateName, group,
                                           defaultValue=defaultValue)


class ALDBRecord(object):
    """Represents an ALDB record."""

    def __init__(self, memory, control_flags, group, address,
                 data1, data2, data3):
        """Initialze the ALDBRecord class."""
        self._memoryLocation = memory
        self._address = Address(address)
        self._group = group
        self._data1 = data1
        self._data2 = data2
        self._data3 = data3
        self._control_flags = ControlFlags.create_from_byte(control_flags)

    def __str__(self):
        """Return the string representation of an ALDB record."""
        props = self._record_properties()
        msgstr = "{"
        first = True
        for prop in props:
            if not first:
                msgstr = '{}, '.format(msgstr)
            for key, val in prop.items():
                if isinstance(val, Address):
                    msgstr = "{}'{}': {}".format(msgstr, key, val.human)
                elif key == 'memory':
                    msgstr = "{}'{}': 0x{:04x}".format(msgstr, key, val)
                elif isinstance(val, int):
                    msgstr = "{}'{}': 0x{:02x}".format(msgstr, key, val)
                else:
                    msgstr = "{}'{}': {}".format(msgstr, key, val)
            first = False
        msgstr = "{}{}".format(msgstr, '}')
        return msgstr

    @property
    def mem_addr(self):
        """Return the memory address of the database record."""
        return self._memoryLocation

    @property
    def memhi(self):
        """Return the memory address MSB."""
        return self._memoryLocation >> 8

    @property
    def memlo(self):
        """Return the memory address LSB."""
        return self._memoryLocation & 0xff

    @property
    def address(self):
        """Return the address of the device the record points to."""
        return self._address

    @property
    def group(self):
        """Return the group the record responds to."""
        return self._group

    @property
    def control_flags(self):
        """Return the record control flags."""
        return self._control_flags

    @property
    def data1(self):
        """Return the data1 field of the ALDB record."""
        return self._data1

    @property
    def data2(self):
        """Return the data2 field of the ALDB record."""
        return self._data2

    @property
    def data3(self):
        """Return the data3 field of the ALDB record."""
        return self._data3

    @staticmethod
    def create_from_userdata(userdata):
        """Create ALDB Record from the userdata dictionary."""
        memhi = userdata.get('d3')
        memlo = userdata.get('d4')
        memory = memhi << 8 | memlo
        control_flags = userdata.get('d6')
        group = userdata.get('d7')
        addrhi = userdata.get('d8')
        addrmed = userdata.get('d9')
        addrlo = userdata.get('d10')
        addr = Address(bytearray([addrhi, addrmed, addrlo]))
        data1 = userdata.get('d11')
        data2 = userdata.get('d12')
        data3 = userdata.get('d13')
        return ALDBRecord(memory, control_flags, group, addr,
                          data1, data2, data3)

    def to_userdata(self):
        """Return a Userdata dictionary."""
        userdata = Userdata({'d3': self.memhi,
                             'd4': self.memlo,
                             'd6': self.control_flags,
                             'd7': self.group,
                             'd8': self.address[2],
                             'd9': self.address[1],
                             'd10': self.address[0],
                             'd11': self.data1,
                             'd12': self.data2,
                             'd13': self.data3})
        return userdata

    def _record_properties(self):
        if self._control_flags.is_controller:
            mode = 'C'
        else:
            mode = 'R'
        rec = [{'memory': self._memoryLocation},
               {'inuse': self._control_flags.is_in_use},
               {'mode':  mode},
               {'highwater': self._control_flags.is_high_water_mark},
               {'group': self.group},
               {'address': self.address},
               {'data1': self.data1},
               {'data2': self.data2},
               {'data3': self.data3}]
        return rec


class ControlFlags(object):
    """Represents a ControlFlag byte of an ALDB record."""

    def __init__(self, in_use, controller, used_before, bit5=0, bit4=0):
        """Initialize the ControlFlags Class."""
        self._in_use = bool(in_use)
        self._controller = bool(controller)
        self._used_before = bool(used_before)
        self._bit5 = bool(bit5)
        self._bit4 = bool(bit4)
        self.log = logging.getLogger()

    @property
    def is_in_use(self):
        """Returns True if the record is in use."""
        return self._in_use

    @property
    def is_available(self):
        """Returns True if the recored is availabe for use."""
        return not self._in_use

    @property
    def is_controller(self):
        """Returns True if the device is a controller."""
        return self._controller

    @property
    def is_responder(self):
        """Returns True if the device is a responder."""
        return not self._controller

    @property
    def is_high_water_mark(self):
        """Returns True if this is the last record."""
        return not self._used_before

    @property
    def is_used_before(self):
        """Returns True if this is not the last record."""
        return self._used_before

    @property
    def byte(self):
        """Returns a byte representation of ControlFlags."""
        flags = int(self._in_use) << 7 \
            | int(self._controller) << 6 \
            | int(self._bit5) << 5 \
            | int(self._bit4) << 4 \
            | int(self._used_before) << 1
        return flags

    @staticmethod
    def create_from_byte(control_flags):
        """Create a ControlFlags class from a control flags byte."""
        in_use = bool(control_flags & 1 << 7)
        controller = bool(control_flags & 1 << 6)
        bit5 = bool(control_flags & 1 << 5)
        bit4 = bool(control_flags & 1 << 4)
        used_before = bool(control_flags & 1 << 1)
        flags = ControlFlags(in_use, controller, used_before,
                             bit5=bit5, bit4=bit4)
        return flags


class ALDBStatus(Enum):
    """All-Link Database load status."""
    EMPTY = 0
    LOADING = 1
    LOADED = 2
    FAILED = 3
    PARTIAL = 4


class ALDBVersion(Enum):
    """All-Link Database version."""
    Null = 0,
    v1 = 1,
    v2 = 2,
    v2cs = 20


LoadAction = collections.namedtuple('LoadAction', 'mem_addr rec_count retries')


class ALDB(object):
    """Represents a device All-Link database."""

    def __init__(self, send_method, loop, address,
                 version=ALDBVersion.v2, mem_addr=0x0000):
        """Instantiate the ALL-Link Database object."""
        self.log = logging.getLogger(__name__)
        self._records = {}
        self._status = ALDBStatus.EMPTY
        self._prior_status = self._status
        self._version = version

        self._send_method = send_method
        self._loop = loop
        self._address = address
        self._mem_addr = mem_addr

        self._rec_mgr_lock = asyncio.Lock(loop=self._loop)
        self._load_action = LoadAction(0, 0, 0)
        self._cb_aldb_loaded = []

    def __len__(self):
        """Return the number of devices in the ALDB."""
        return len(self._records)

    def __iter__(self):
        """Iterate through each ALDB device record."""
        keys = list(self._records.keys())
        keys.sort(reverse=True)
        for key in keys:
            yield key

    def __getitem__(self, mem_addr):
        """Fetch a device from the ALDB."""
        return self._records.get(mem_addr)

    def __setitem__(self, mem_addr, record):
        """Add or Update a device in the ALDB."""
        if not isinstance(record, ALDBRecord):
            raise ValueError

        self._records[mem_addr] = record

    def __repr__(self):
        """Human representation of a device from the ALDB."""
        attrs = vars(self)
        return ', '.join("%s: %r" % item for item in attrs.items())

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, val: ALDBStatus):
        self._status = val

    @property
    def version(self):
        return self._version

    def pop(self, key):
        """Pop and remove an item from the ALDB."""
        return self._records.pop(key)

    def add_loaded_callback(self, callback):
        """Add a callback to be run when the ALDB load is complete."""
        if callback not in self._cb_aldb_loaded:
            self._cb_aldb_loaded.append(callback)

    @asyncio.coroutine
    def load(self, mem_addr=0x0000, rec_count=0, retry=0):
        """Read the device database and load."""
        if self._version == ALDBVersion.Null:
            self._status = ALDBStatus.LOADED
            self.log.debug('Device has no ALDB')
        else:
            self._status = ALDBStatus.LOADING
            self.log.debug('Tring to lock from load')
            yield from self._rec_mgr_lock
            self.log.debug('load yielded lock')

            mem_hi = mem_addr >> 8
            mem_lo = mem_addr & 0xff
            log_output = 'ALDB read'
            max_retries = 0
            if rec_count:
                max_retries = ALDB_RECORD_RETRIES
                if mem_addr == 0x0000:
                    log_output = '{:s} first record'.format(log_output)
                else:
                    log_output = '{:s} record {:04x}'.format(log_output,
                                                             mem_addr)
            else:
                max_retries = ALDB_ALL_RECORD_RETRIES
                log_output = '{:s} all records'.format(log_output)

            if retry:
                log_output = '{:s} retry {:d} of {:d}'.format(log_output,
                                                              retry,
                                                              max_retries)
            self.log.info(log_output)
            chksum = 0xff - ((0x2f+mem_hi+mem_lo+1) & 0xff) + 1
            userdata = Userdata({'d1': 0,
                                 'd2': 0,
                                 'd3': mem_hi,
                                 'd4': mem_lo,
                                 'd5': rec_count,
                                 'd14': chksum})
            msg = ExtendedSend(self._address,
                               COMMAND_EXTENDED_READ_WRITE_ALDB_0X2F_0X00,
                               userdata=userdata)
            self._send_method(msg, self._handle_read_aldb_ack, True)

            if not self._load_action:
                self._set_load_action(mem_addr, rec_count, -1, False)

    def get(self, mem_addr):
        """Return an All-Link record at a memory address."""
        return self._records.get(mem_addr)

    def clear(self):
        """Remove all records."""
        self._records.clear()

    def write_record(self, mem_addr: int, mode: str, group: int, target,
                     data1=0x00, data2=0x00, data3=0x00):
        """Write an All-Link database record."""
        if not (self._have_first_record() and self._have_last_record()):
            self.log.error('Must load the ALDB before writing to it')
        else:
            self._prior_status = self._status
            self._status = ALDBStatus.LOADING
            mem_hi = mem_addr >> 8
            mem_lo = mem_addr & 0xff
            controller = True if mode == 'c' else False
            control_flag = ControlFlags(True, controller, True, False, False)
            addr = Address(target)
            addr_lo = addr.bytes[0]
            addr_mid = addr.bytes[1]
            addr_hi = addr.bytes[2]
            chksum = 0xff - ((0x2f + 0x02 + mem_hi + mem_lo + 0x08 +
                              control_flag.byte +
                              addr_lo + addr_mid + addr_hi +
                              group + data1 + data2 + data3) & 0xff) + 1
            userdata = Userdata({'d1': 0,
                                 'd2': 0x02,
                                 'd3': mem_hi,
                                 'd4': mem_lo,
                                 'd5': 0x08,
                                 'd6': control_flag.byte,
                                 'd7': group,
                                 'd8': addr_lo,
                                 'd9': addr_mid,
                                 'd10': addr_hi,
                                 'd11': data1,
                                 'd12': data2,
                                 'd13': data3,
                                 'd14': chksum})
            msg = ExtendedSend(self._address,
                               COMMAND_EXTENDED_READ_WRITE_ALDB_0X2F_0X00,
                               userdata=userdata)
            self.log.info('writing message %s', msg)
            self._send_method(msg, self._handle_write_aldb_ack, True)
            self._load_action = LoadAction(mem_addr, 1,  0)

    def del_record(self, mem_addr: int):
        """Write an All-Link database record."""
        record = self._records.get(mem_addr)
        if not record:
            self.log.error('Must load the ALDB record before deleting it')
        else:
            self._prior_status = self._status
            self._status = ALDBStatus.LOADING
            mem_hi = mem_addr >> 8
            mem_lo = mem_addr & 0xff
            controller = record.control_flags.is_controller
            control_flag = ControlFlags(False, controller, True, False, False)
            addr = record.address
            addr_lo = addr.bytes[0]
            addr_mid = addr.bytes[1]
            addr_hi = addr.bytes[2]
            group = record.group
            data1 = record.data1
            data2 = record.data2
            data3 = record.data3
            chksum = 0xff - ((0x2f + 0x02 + mem_hi + mem_lo + 0x08 +
                              control_flag.byte +
                              addr_lo + addr_mid + addr_hi +
                              group + data1 + data2 + data3) & 0xff) + 1
            userdata = Userdata({'d1': 0,
                                 'd2': 0x02,
                                 'd3': mem_hi,
                                 'd4': mem_lo,
                                 'd5': 0x08,
                                 'd6': control_flag.byte,
                                 'd7': group,
                                 'd8': addr_lo,
                                 'd9': addr_mid,
                                 'd10': addr_hi,
                                 'd11': data1,
                                 'd12': data2,
                                 'd13': data3,
                                 'd14': chksum})
            msg = ExtendedSend(self._address,
                               COMMAND_EXTENDED_READ_WRITE_ALDB_0X2F_0X00,
                               userdata=userdata)
            self.log.info('writing message %s', msg)
            self._send_method(msg, self._handle_write_aldb_ack, True)
            self._load_action = LoadAction(mem_addr, 1,  0)

    def find_matching_link(self, mode, group, addr):
        """Finds a matching link in the current device.

        Mode: r | c is the mode of the link in the linked device
              This method will search for a corresponding link in the
              reverse direction.
        group: All-Link group number
        addr:  Inteon address of the linked device
        """
        found_rec = None
        mode_test = None
        if mode.lower() in ['c', 'r']:
            link_group = int(group)
            link_addr = Address(addr)
            for mem_addr in self:
                rec = self[mem_addr]
                if mode.lower() == 'r':
                    mode_test = rec.control_flags.is_controller
                else:
                    mode_test = rec.control_flags.is_responder
                if (mode_test and
                        rec.group == link_group and
                        rec.address == link_addr):
                    found_rec = rec
        return found_rec

    def record_received(self, msg):
        """Handle ALDB record received from device."""
        release_lock = False
        userdata = msg.userdata
        rec = ALDBRecord.create_from_userdata(userdata)
        self._records[rec.mem_addr] = rec

        self.log.debug('ALDB Record: %s', rec)

        rec_count = self._load_action.rec_count
        if rec_count == 1 or self._have_all_records():
            release_lock = True

        if self._is_first_record(rec):
            self._mem_addr = rec.mem_addr

        if release_lock and self._rec_mgr_lock.locked():
            self.log.debug('Releasing lock because record received')
            self._rec_mgr_lock.release()

    def load_saved_records(self, status, records):
        if isinstance(status, ALDBStatus):
            self._status = status
        else:
            self._status = ALDBStatus(status)
        for mem_addr in records:
            rec = records[mem_addr]
            control_flags = int(rec.get('control_flags', 0))
            group = int(rec.get('group', 0))
            rec_addr = rec.get('address', '000000')
            data1 = int(rec.get('data1', 0))
            data2 = int(rec.get('data2', 0))
            data3 = int(rec.get('data3', 0))
            self[int(mem_addr)] = ALDBRecord(int(mem_addr), control_flags,
                                             group, rec_addr,
                                             data1, data2, data3)
        if self._status == ALDBStatus.LOADED:
            keys = list(self._records.keys())
            keys.sort(reverse=True)
            first_key = keys[0]
            self._mem_addr == first_key

    def _handle_read_aldb_ack(self, msg):
        self.log.debug('Read ALDB directACK received or timeout reached')
        asyncio.ensure_future(self._read_timeout_manager(), loop=self._loop)

    def _handle_write_aldb_ack(self, msg):
        if msg:
            self.log.info('Device confirmed ALDB message write')
            try:
                self._records.pop(self._load_action.mem_addr)
            except KeyError:
                pass
            asyncio.ensure_future(self.load(self._load_action.mem_addr, 1, 0),
                                  loop=self._loop)
        else:
            self.log.info('Device did not confirm ALDB message write')
            self._status = self._prior_status

    @asyncio.coroutine
    def _read_timeout_manager(self):
        self.log.debug('_read_timeout_manager started.')
        read_complete = False
        rec_count = self._load_action.rec_count
        timeout = ALDB_RECORD_TIMEOUT if rec_count else ALDB_ALL_RECORD_TIMEOUT

        try:
            with async_timeout.timeout(timeout + self._load_action.retries):
                self.log.debug('Tring to get lock in _read_timeout_manager.')
                yield from self._rec_mgr_lock
                self.log.debug('_read_timeout_manager lock yielded.')
                read_complete = True
        except asyncio.TimeoutError:
            if (self._load_action.rec_count and
                    self._load_action.mem_addr == 0x0000 and
                    self._have_first_record()):
                read_complete = False
            elif self.get(self._load_action.mem_addr):
                read_complete = True
            else:
                self.log.debug('ALDB record response timeout.')
                read_complete = False

        self._set_load_action(self._load_action.mem_addr,
                              self._load_action.rec_count,
                              self._load_action.retries,
                              read_complete)

        mem_addr = self._load_action.mem_addr
        rec_count = self._load_action.rec_count
        retries = self._load_action.retries

        if self._rec_mgr_lock.locked():
            self._rec_mgr_lock.release()

        if mem_addr is not None:
            asyncio.ensure_future(self.load(mem_addr, rec_count, retries),
                                  loop=self._loop)
        elif read_complete or self._have_all_records():
            self.log.info('ALDB load complete for device %s', self._address)
            self._load_finished(ALDBStatus.LOADED)
        else:
            if self._records:
                self.log.warning('ALDB partially loaded for device %s',
                                 self._address)
                self._load_finished(ALDBStatus.PARTIAL)
            else:
                self.log.warning('ALDB failed to load for device %s',
                                 self._address)
                self._load_finished(ALDBStatus.FAILED)

    def _load_finished(self, status):
            self._status = status
            while self._cb_aldb_loaded:
                self.log.info('Calling aldb loaded callback')
                callback = self._cb_aldb_loaded.pop()
                if callback:
                    callback()
            self._load_action = LoadAction(0, 0, 0)

    def _set_load_action(self, mem_addr, rec_count, retries,
                         read_complete=False):
        """Calculates the next record to read.

        If the last record was successful and one record was being read then
        look for the next record until we get to the high water mark.

        If the last read was successful and all records were being read then
        look for the first record.

        if the last read was unsuccessful and one record was being read then
        repeat the last read until max retries

        If the last read was unsuccessful and all records were being read then
        repeat the last read until max retries or look for the first record.
        """
        if self._have_all_records():
            mem_addr = None
            rec_count = 0
            retries = 0
        elif read_complete:
            retries = 0
            if rec_count:
                mem_addr = self._next_address(mem_addr)
            else:
                mem_addr = self._next_address(mem_addr)
                rec_count = 1
                retries = 0
        elif rec_count and retries < ALDB_RECORD_RETRIES:
            retries = retries + 1
        elif not rec_count and retries < ALDB_ALL_RECORD_RETRIES:
            retries = retries + 1
        elif not rec_count and retries >= ALDB_ALL_RECORD_RETRIES:
            mem_addr = self._next_address(mem_addr)
            rec_count = 1
            retries = 0
        else:
            mem_addr = None
            rec_count = 0
            retries = 0

        self._load_action = LoadAction(mem_addr, rec_count,  retries)
        if mem_addr is not None:
            self.log.debug('Load action: addr: %04x rec_count: %d retries: %d',
                           self._load_action.mem_addr,
                           self._load_action.rec_count,
                           self._load_action.retries)

    def _next_address(self, mem_addr):
        if self._have_first_record() and mem_addr == 0x0000:
            mem_addr = self._mem_addr
        while self.get(mem_addr):
            if self.get(mem_addr).control_flags.is_high_water_mark:
                mem_addr = None
            else:
                mem_addr = mem_addr - 8
        return mem_addr

    def _have_first_record(self):
        have_first_record = False
        if self._records:
            keys = list(self._records.keys())
            keys.sort(reverse=True)
            first_key = int(keys[0])
            if first_key == self._mem_addr:
                have_first_record = True
            elif first_key == 0x0fff:
                self._mem_addr = first_key
                have_first_record = True
        return have_first_record

    def _have_last_record(self):
        have_last_record = False
        if self._records:
            keys = list(self._records.keys())
            keys.sort(reverse=False)
            last_key = keys[0]
            if self[last_key].control_flags.is_high_water_mark:
                have_last_record = True
        return have_last_record

    def _is_first_record(self, rec):
        is_first_record = False
        mem_addr = self._load_action.mem_addr
        rec_count = self._load_action.rec_count
        if rec_count and mem_addr == 0x0000:
            is_first_record = True
        elif self._mem_addr == rec.mem_addr:
            is_first_record = True
        elif (self._mem_addr == 0x0000 and
              rec.mem_addr == 0x0fff):
            is_first_record = True
        return is_first_record

    def _have_all_records(self):
        have_all_recs = False
        if self._have_first_record() and self._have_last_record():
            prev_mem_addr = None
            for mem_addr in self:
                if not prev_mem_addr:
                    prev_mem_addr = mem_addr
                elif mem_addr == (prev_mem_addr - 8):
                    if self[mem_addr].control_flags.is_high_water_mark:
                        have_all_recs = True
                    else:
                        prev_mem_addr = mem_addr
                else:
                    have_all_recs = False
                    break
        return have_all_recs



"""INSTEON Message All-Link Record Response."""
from insteonplm.messages.message import Message
from insteonplm.constants import (MESSAGE_ALL_LINK_RECORD_RESPONSE_0X57,
                                  MESSAGE_ALL_LINK_RECORD_RESPONSE_SIZE)
from insteonplm.address import Address


class AllLinkRecordResponse(Message):
    """INSTEON ALL-Link Record Response.

    Message type 0x57
    """

    _code = MESSAGE_ALL_LINK_RECORD_RESPONSE_0X57
    _sendSize = MESSAGE_ALL_LINK_RECORD_RESPONSE_SIZE
    _receivedSize = MESSAGE_ALL_LINK_RECORD_RESPONSE_SIZE
    _description = 'INSTEON ALL-Link Record Response'

    def __init__(self, flags, group, address, linkdata1, linkdata2, linkdata3):
        """Initalize the AllLinkRecordResponse Class."""
        self._controlFlags = flags
        self._group = group
        self._address = Address(address)
        self._linkdata1 = linkdata1
        self._linkdata2 = linkdata2
        self._linkdata3 = linkdata3

    @classmethod
    def from_raw_message(cls, rawmessage):
        """Create message from raw byte stream."""
        return AllLinkRecordResponse(rawmessage[2],
                                     rawmessage[3],
                                     rawmessage[4:7],
                                     rawmessage[7],
                                     rawmessage[8],
                                     rawmessage[9])

    @property
    def controlFlags(self):
        """Return the link record control flags."""
        return self._controlFlags

    @property
    def group(self):
        """Return the link record group."""
        return self._group

    @property
    def address(self):
        """Return the device address."""
        return self._address

    @property
    def linkdata1(self):
        """Return the first link data field."""
        return self._linkdata1

    @property
    def linkdata2(self):
        """Return the second link data field."""
        return self._linkdata2

    @property
    def linkdata3(self):
        """Return the third link data field."""
        return self._linkdata3

    @property
    def isRecordinuse(self):
        """Test if the link record is in use."""
        return ((self._controlFlags & self.CONTROL_FLAG_RECORD_IN_USE) ==
                self.CONTROL_FLAG_RECORD_IN_USE)

    @property
    def isController(self):
        """Test if the link group is a controller."""
        return ((self._controlFlags & self.CONTROL_FLAG_CONTROLLER) ==
                self.CONTROL_FLAG_CONTROLER)

    @property
    def isSlave(self):
        """Test if the link group is a slave or responder."""
        return not self.iscontroller

    def _message_properties(self):
        return [{'controlFlags': self._controlFlags},
                {'group': self._group},
                {'address': self._address},
                {'linkdata1': self._linkdata1},
                {'linkdata2': self._linkdata2},
                {'linkdata3': self._linkdata3}]




"""Helper objects for maintaining PLM state and interfaces."""
import logging
import binascii
import insteonplm.utils

__all__ = ('Address')
_LOGGER = logging.getLogger(__name__)


class Address(object):
    """Datatype definition for INSTEON device address handling."""

    def __init__(self, addr):
        """Create an Address object."""
        self._is_x10 = False
        self.addr = self._normalize(addr)

    def __repr__(self):
        """Representation of the Address object."""
        return self.id

    def __str__(self):
        """String representation of the Address object."""
        return self.id

    def __eq__(self, other):
        """Test for equality."""
        equals = False
        if hasattr(other, 'addr'):
            equals = self.addr == other.addr
        return equals

    def __ne__(self, other):
        """Test for not equals."""
        not_equals = True
        if hasattr(other, 'addr'):
            not_equals = self.addr != other.addr
        return not_equals

    def __lt__(self, other):
        """Test for less than."""
        if isinstance(other, Address):
            return str(self) < str(other)
        else:
            raise TypeError

    def __gt__(self, other):
        """Test for greater than."""
        if isinstance(other, Address):
            return str(self) > str(other)
        else:
            raise TypeError

    def __hash__(self):
        """Create a hash code for the Address object."""
        return hash(self.id)

    def matches_pattern(self, other):
        """Test Address object matches the pattern of another  object."""
        matches = False
        if hasattr(other, 'addr'):
            if self.addr is None or other.addr is None:
                matches = True
            else:
                matches = self.addr == other.addr
        return matches

    def _normalize(self, addr):
        """Take any format of address and turn it into a hex string."""
        normalize = None
        if isinstance(addr, Address):
            normalize = addr.addr
            self._is_x10 = addr.is_x10

        elif isinstance(addr, bytearray):
            normalize = binascii.unhexlify(binascii.hexlify(addr).decode())

        elif isinstance(addr, bytes):
            normalize = addr

        elif isinstance(addr, str):
            addr = addr.replace('.', '')
            addr = addr[0:6]
            if addr[0:3].lower() == 'x10':
                x10_addr = Address.x10(addr[3:4], int(addr[4:6]))
                normalize = x10_addr.addr
                self._is_x10 = True
            else:
                normalize = binascii.unhexlify(addr.lower())

        elif addr is None:
            normalize = None

        else:
            _LOGGER.warning('Address class init with unknown type %s: %r',
                            type(addr), addr)
        return normalize

    @property
    def human(self):
        """Emit the address in human-readible format (AA.BB.CC)."""
        addrstr = '00.00.00'
        if self.addr:
            if self._is_x10:
                housecode_byte = self.addr[1]
                housecode = insteonplm.utils.byte_to_housecode(housecode_byte)
                unitcode_byte = self.addr[2]
                unitcode = insteonplm.utils.byte_to_unitcode(unitcode_byte)
                addrstr = 'X10.{}.{:02d}'.format(housecode.upper(), unitcode)
            else:
                addrstr = '{}.{}.{}'.format(self.hex[0:2],
                                            self.hex[2:4],
                                            self.hex[4:6]).upper()
        return addrstr

    @property
    def hex(self):
        """Emit the address in bare hex format (aabbcc)."""
        addrstr = '000000'
        if self.addr is not None:
            addrstr = binascii.hexlify(self.addr).decode()
        return addrstr

    @property
    def bytes(self):
        """Emit the address in bytes format."""
        addrbyte = b'\x00\x00\x00'
        if self.addr is not None:
            addrbyte = self.addr
        return addrbyte

    @property
    def id(self):
        """Return the ID of the device address."""
        id = ''
        if self._is_x10:
            id = 'x10{}{:02d}'.format(self.x10_housecode, self.x10_unitcode)
        else:
            id = self.hex
        return id

    @property
    def is_x10(self):
        """Test if this is an X10 address."""
        return self._is_x10

    @is_x10.setter
    def is_x10(self, val: bool):
        """Set if this is an X10 address."""
        self._is_x10 = val

    @property
    def x10_housecode_byte(self):
        """Emit the X10 house code byte value."""
        housecode = None
        if self.is_x10:
            housecode = self.addr[1]
        return housecode

    @property
    def x10_unitcode_byte(self):
        """Emit the X10 unit code byte value."""
        unitcode = None
        if self.is_x10:
            unitcode = self.addr[2]
        return unitcode

    @property
    def x10_housecode(self):
        """Emit the X10 house code."""
        housecode = None
        if self.is_x10:
            housecode = insteonplm.utils.byte_to_housecode(self.addr[1])
        return housecode

    @property
    def x10_unitcode(self):
        """Emit the X10 unit code."""
        unitcode = None
        if self.is_x10:
            unitcode = insteonplm.utils.byte_to_unitcode(self.addr[2])
        return unitcode

    @classmethod
    def x10(cls, housecode, unitcode):
        """Create an X10 device address."""
        if housecode.lower() in ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h',
                                 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p']:
            byte_housecode = insteonplm.utils.housecode_to_byte(housecode)
        else:
            if isinstance(housecode, str):
                _LOGGER.error('X10 house code error: %s', housecode)
            else:
                _LOGGER.error('X10 house code is not a string')
            raise ValueError

        # 20, 21 and 22 for All Units Off, All Lights On and All Lights Off
        # 'fake' units
        if unitcode in range(1, 17) or unitcode in range(20, 23):
            byte_unitcode = insteonplm.utils.unitcode_to_byte(unitcode)
        else:
            if isinstance(unitcode, int):
                _LOGGER.error('X10 unit code error: %d', unitcode)
            else:
                _LOGGER.error('X10 unit code is not a integer 1 - 16')
            raise ValueError

        addr = Address(bytearray([0x00, byte_housecode, byte_unitcode]))
        addr.is_x10 = True
        return addr



"""INSTEON Message Get IM Configuration."""
from insteonplm.constants import (MESSAGE_GET_IM_CONFIGURATION_0X73,
                                  MESSAGE_GET_IM_CONFIGURATION_SIZE,
                                  MESSAGE_GET_IM_CONFIGURATION_RECEIVED_SIZE,
                                  MESSAGE_ACK,
                                  MESSAGE_NAK)
from insteonplm.messages.message import Message


class GetImConfiguration(Message):
    """Insteon Get IM Configuration Message.

    Message type 0x62
    """

    _code = MESSAGE_GET_IM_CONFIGURATION_0X73
    _sendSize = MESSAGE_GET_IM_CONFIGURATION_SIZE
    _receivedSize = MESSAGE_GET_IM_CONFIGURATION_RECEIVED_SIZE
    _description = 'Insteon Get IM Configuration Message'

    def __init__(self, flags=None, acknak=None):
        """Initialize the GetImConfiguration Class."""
        self._imConfigurationFlags = flags
        self._spare1 = None
        self._spare2 = None
        self._acknak = self._setacknak(acknak)

    @classmethod
    def from_raw_message(cls, rawmesssage):
        """Create message from a raw byte stream."""
        return GetImConfiguration(rawmesssage[2],
                                  rawmesssage[5])

    @property
    def imConfigurationFlags(self):
        """Return the IM configuration flags."""
        return self._imConfigurationFlags

    @property
    def acknak(self):
        """Return the ACK/NAK byte."""
        return self._acknak

    @property
    def isack(self):
        """Test if this is an ACK message."""
        return self._acknak is not None and self._acknak == MESSAGE_ACK

    @property
    def isnak(self):
        """Test if this is a NAK message."""
        return self._acknak is not None and self._acknak == MESSAGE_NAK

    def _message_properties(self):
        if self._imConfigurationFlags is not None:
            self._spare1 = 0x00
            self._spare2 = 0x00
        else:
            self._spare1 = None
            self._spare2 = None
        return [{'imConfigurationFlags': self._imConfigurationFlags},
                {'spare1': self._spare1},
                {'spare2': self._spare2},
                {'acknak': self.acknak}]



"""INSTEON Message All-LinkCleanup Failup Report."""
from insteonplm.messages.message import Message
from insteonplm.constants import (MESSAGE_ALL_LINK_CEANUP_FAILURE_REPORT_0X56,
                                  MESSAGE_ALL_LINK_CEANUP_FAILURE_REPORT_SIZE)
from insteonplm.address import Address


class AllLinkCleanupFailureReport(Message):
    """INSTEON All-Link Failure Report Message.

    Message type 0x56
    """

    _code = MESSAGE_ALL_LINK_CEANUP_FAILURE_REPORT_0X56
    _sendSize = MESSAGE_ALL_LINK_CEANUP_FAILURE_REPORT_SIZE
    _receivedSize = MESSAGE_ALL_LINK_CEANUP_FAILURE_REPORT_SIZE
    _description = 'INSTEON All-Link Failure Report Message'

    def __init__(self, group, address):
        """Initialize the AllLinkCleanupFailureReport Class."""
        self._group = group
        self._address = Address(address)
        self._failedFlag = 0x01

    @classmethod
    def from_raw_message(cls, rawmessage):
        """Create message from a raw byte stream."""
        return AllLinkCleanupFailureReport(rawmessage[3], rawmessage[4:7])

    @property
    def group(self):
        """Return the All-Link Group."""
        return self._group

    @property
    def address(self):
        """Return the device address."""
        return self._address

    def _message_properties(self):
        return [{'failedFlag': self._failedFlag},
                {'group': self._group},
                {'address': self._address}]



"""Test message callbacks."""

import logging
from insteonplm.messagecallback import MessageCallback
from insteonplm.constants import (COMMAND_LIGHT_ON_0X11_NONE,
                                  MESSAGE_ACK,
                                  MESSAGE_NAK)
from insteonplm.messages.standardReceive import StandardReceive
from insteonplm.messages.standardSend import StandardSend
from insteonplm.messages.extendedReceive import ExtendedReceive
from insteonplm.messages.extendedSend import ExtendedSend
from insteonplm.messages.allLinkRecordResponse import AllLinkRecordResponse
from insteonplm.messages.getIMInfo import GetImInfo
from insteonplm.messages.getNextAllLinkRecord import GetNextAllLinkRecord
from insteonplm.messages.userdata import Userdata
from .mockCallbacks import MockCallbacks


_LOG = logging.getLogger(__name__)


def test_messagecallback_basic():
    """Test messagecallback basic."""
    callbacks = MessageCallback()
    callbacktest1 = "test callback 1"

    msg_template = StandardReceive.template(
        commandtuple=COMMAND_LIGHT_ON_0X11_NONE, flags=0x80)
    callbacks[msg_template] = callbacktest1

    msg = StandardReceive('1a2b3c', '4d5e6f', COMMAND_LIGHT_ON_0X11_NONE,
                          cmd2=0xff, flags=0x80)

    callback1 = callbacks.get_callbacks_from_message(msg)

    assert len(callback1) == 1
    assert callback1[0] == callbacktest1


def test_messagecallback_msg():
    """Test messagecallback msg."""
    callbacks = MessageCallback()
    callbacktest = "test callback"
    address = '1a2b3c'
    target = '4d5e6f'

    template_on = StandardReceive.template(
        commandtuple=COMMAND_LIGHT_ON_0X11_NONE)
    callbacks.add(template_on, callbacktest)
    msg1 = StandardReceive(address, target, COMMAND_LIGHT_ON_0X11_NONE,
                           cmd2=0x00)
    msg2 = StandardReceive(address, target, COMMAND_LIGHT_ON_0X11_NONE,
                           cmd2=0xff)

    callback1 = callbacks.get_callbacks_from_message(msg1)
    callback2 = callbacks.get_callbacks_from_message(msg2)

    assert callback1[0] == callbacktest
    assert callback2[0] == callbacktest


def test_messagecallback_acknak():
    """Test messagecallback acknak."""
    callbacks = MessageCallback()
    callbacktest1 = "test callback 1"
    callbacktest2 = "test callback 2"
    callbacktest3 = "test callback 3"
    callbacktest4 = "test callback 4"
    address = '1a2b3c'

    template_address = StandardSend.template(address=address)
    template_address_ack = StandardSend.template(address=address,
                                                 acknak=MESSAGE_ACK)
    template_empty = StandardSend.template()
    template_nak = StandardSend.template(acknak=MESSAGE_NAK)
    callbacks.add(template_address, callbacktest1)
    callbacks.add(template_address_ack, callbacktest2)
    callbacks.add(template_empty, callbacktest3)
    callbacks.add(template_nak, callbacktest4)

    msg1 = StandardSend(address, COMMAND_LIGHT_ON_0X11_NONE, cmd2=0xcd)
    msg2 = StandardSend('222222', COMMAND_LIGHT_ON_0X11_NONE, cmd2=0xff)
    msg3 = StandardSend('333333', COMMAND_LIGHT_ON_0X11_NONE, cmd2=0xff,
                        acknak=MESSAGE_NAK)
    msg4 = StandardSend('444444', COMMAND_LIGHT_ON_0X11_NONE, cmd2=0xff,
                        acknak=MESSAGE_ACK)

    _LOG.debug('Getting callbacks for message 1')
    callback1 = callbacks.get_callbacks_from_message(msg1)
    _LOG.debug('Getting callbacks for message 2')
    callback2 = callbacks.get_callbacks_from_message(msg2)
    _LOG.debug('Getting callbacks for message 3')
    callback3 = callbacks.get_callbacks_from_message(msg3)
    _LOG.debug('Getting callbacks for message 4')
    callback4 = callbacks.get_callbacks_from_message(msg4)

    assert len(callback1) == 4
    assert len(callback2) == 2
    assert len(callback3) == 2
    assert len(callback4) == 1


def test_message_callback_extended():
    """Test message callback extended."""
    callbacks = MessageCallback()
    callbacktest = "test callback"
    address = '1a2b3c'
    target = '4d5e6f'

    template_ext_on = ExtendedReceive.template(
        commandtuple=COMMAND_LIGHT_ON_0X11_NONE,
        userdata=Userdata({'d1': 0x02}))
    callbacks.add(template_ext_on, callbacktest)
    msg1 = ExtendedReceive(address, target, COMMAND_LIGHT_ON_0X11_NONE,
                           Userdata({'d1': 0x02}), cmd2=0xff)
    msg2 = ExtendedReceive(address, target, COMMAND_LIGHT_ON_0X11_NONE,
                           Userdata({'d1': 0x03, 'd2': 0x02}), cmd2=0xff)

    callback1 = callbacks.get_callbacks_from_message(msg1)
    callback2 = callbacks.get_callbacks_from_message(msg2)

    assert callback1[0] == callbacktest
    assert not callback2


def test_delete_callback():
    """Test delete callback."""
    callbacks = MessageCallback()
    callbacktest1 = "test callback 1"
    callbacktest2 = "test callback 2"
    callbacktest3 = "test callback 3"

    callbacks.add(StandardSend.template(), callbacktest1)
    callbacks.add(StandardSend.template(), callbacktest2)
    callbacks.add(StandardSend.template(acknak=MESSAGE_NAK), callbacktest3)

    msg = StandardSend('333333', COMMAND_LIGHT_ON_0X11_NONE, cmd2=0xff,
                       acknak=MESSAGE_NAK)

    callback_list = callbacks.get_callbacks_from_message(msg)
    assert len(callback_list) == 3

    callbacks.remove(StandardSend.template(), callbacktest2)
    callback_list = callbacks.get_callbacks_from_message(msg)
    assert len(callback_list) == 2


def test_misc_messages():
    """Test misc messages."""
    callbacks = MessageCallback()
    callbacktest1 = "test callback 1"
    callbacktest2 = "test callback 2"
    callbacktest3 = "test callback 3"

    msgtemplate1 = AllLinkRecordResponse(None, None, None, None, None, None)
    msgtemplate2 = GetImInfo()
    msgtemplate3 = GetNextAllLinkRecord(acknak=MESSAGE_NAK)
    callbacks.add(msgtemplate1, callbacktest1)
    callbacks.add(msgtemplate2, callbacktest2)
    callbacks.add(msgtemplate3, callbacktest3)

    msg1 = AllLinkRecordResponse(0x00, 0x01, '1a2b3c', 0x01, 0x02, 0x03)
    msg2 = GetImInfo()
    msg3 = GetNextAllLinkRecord(acknak=MESSAGE_ACK)
    msg4 = GetNextAllLinkRecord(acknak=MESSAGE_NAK)

    callback_list1 = callbacks.get_callbacks_from_message(msg1)
    callback_list2 = callbacks.get_callbacks_from_message(msg2)
    callback_list3 = callbacks.get_callbacks_from_message(msg3)
    callback_list4 = callbacks.get_callbacks_from_message(msg4)

    assert callback_list1[0] == callbacktest1
    assert callback_list2[0] == callbacktest2
    assert not callback_list3
    assert callback_list4[0] == callbacktest3


def test_extended_ack():
    """Test extended ack."""
    callbacks = MockCallbacks()
    callbacks.callbackvalue1 = "Callback 1"
    callbacks.callbackvalue2 = "Callback 2"
    message_callbacks = MessageCallback()
    address = '1a2b3c'

    template_ext_ack = ExtendedSend.template(address, acknak=MESSAGE_ACK)
    template_std_ack = StandardSend.template(address, acknak=MESSAGE_ACK)

    message_callbacks.add(template_ext_ack, callbacks.callbackvalue1)
    message_callbacks.add(template_std_ack, callbacks.callbackvalue2)
    extmsg = ExtendedSend(address, COMMAND_LIGHT_ON_0X11_NONE, {'d1': 0x02},
                          cmd2=0xff, acknak=MESSAGE_ACK)
    stdmsg = StandardSend(address, COMMAND_LIGHT_ON_0X11_NONE, cmd2=0xff,
                          acknak=MESSAGE_ACK)
    result1 = message_callbacks.get_callbacks_from_message(extmsg)
    result2 = message_callbacks.get_callbacks_from_message(stdmsg)

    assert result2 == [callbacks.callbackvalue2]
    assert result1 == [callbacks.callbackvalue1]



"""INSTEON Message Get IM Info."""

from insteonplm.messages.message import Message
from insteonplm.constants import (MESSAGE_GET_IM_INFO_0X60,
                                  MESSAGE_GET_IM_INFO_SIZE,
                                  MESSAGE_GET_IM_INFO_RECEIVED_SIZE,
                                  MESSAGE_ACK,
                                  MESSAGE_NAK)
from insteonplm.address import Address


class GetImInfo(Message):
    """INSTEON Get Insteon Modem Info Message.

    Message type 0x60
    """

    _code = MESSAGE_GET_IM_INFO_0X60
    _sendSize = MESSAGE_GET_IM_INFO_SIZE
    _receivedSize = MESSAGE_GET_IM_INFO_RECEIVED_SIZE
    _description = 'INSTEON Get Insteon Modem Info Message Received'

    def __init__(self, address=None, cat=None, subcat=None, firmware=None,
                 acknak=None):
        """Initialize the GetImInfo Class."""
        self._address = Address(address)
        self._category = cat
        self._subcategory = subcat
        self._firmware = firmware
        self._acknak = self._setacknak(acknak)

    @classmethod
    def from_raw_message(cls, rawmessage):
        """Create message from raw byte stream."""
        return GetImInfo(rawmessage[2:5],
                         rawmessage[5],
                         rawmessage[6],
                         rawmessage[7],
                         rawmessage[8])

    @property
    def address(self):
        """Return the device address."""
        return self._address

    @property
    def category(self):
        """Return the IM device category."""
        return self._category

    @property
    def subcategory(self):
        """Return the IM device subcategory."""
        return self._subcategory

    @property
    def firmware(self):
        """Return the IM device firmware version."""
        return self._firmware

    @property
    def acknak(self):
        """Return the ACK/NAK byte."""
        return self._acknak

    @property
    def isack(self):
        """Test if this is an ACK message."""
        return self._acknak is not None and self._acknak == MESSAGE_ACK

    @property
    def isnak(self):
        """Test if this is a NAK message."""
        return self._acknak is not None and self._acknak == MESSAGE_NAK

    def _message_properties(self):
        return [{'address': self._address},
                {'category': self._category},
                {'subcategory': self._subcategory},
                {'firmware': self._firmware},
                {'acknak': self._acknak}]



"""Provides a raw console to test module and demonstrate usage."""
import argparse
import asyncio
import binascii
import logging
import string
import sys

import insteonplm
from insteonplm.address import Address
from insteonplm.devices import Device, ALDBStatus

__all__ = ('Tools', 'monitor', 'interactive')
_LOGGING = logging.getLogger()
PROMPT = 'insteonplm: '
INTRO = ('INSTEON PLM interactive command processor.\n'
         'Type `help` for a list of commands.\n\n'
         'This command line tool is still in development.\n'
         '!!!! Please be VERY careful with the `write_aldb` command !!!!!!!')
ALLOWEDCHARS = string.ascii_letters + string.digits + '_'


class Tools():
    """Set of tools to support utility programs."""
    def __init__(self, loop, args=None):
        """Create Tools class."""
        # common variables
        self.loop = loop
        self.plm = insteonplm.PLM()
        self.logfile = None
        self.workdir = None
        self.loglevel = logging.INFO

        # connection variables
        self.device = args.device
        self.username = None
        self.password = None
        self.host = None
        self.port = None

        # all-link variables
        self.address = None
        self.linkcode = None
        self.group = None
        self.wait_time = 10

        self.aldb_load_lock = asyncio.Lock(loop=loop)

        if args:
            if args.verbose:
                self.loglevel = logging.DEBUG
            else:
                self.loglevel = logging.INFO

            if hasattr(args, "workdir"):
                self.workdir = args.workdir
            if hasattr(args, "logfile"):
                self.logfile = args.logfile
            if hasattr(args, 'address'):
                self.address = args.address
            if hasattr(args, 'group'):
                self.group = int(args.group)
            if hasattr(args, 'linkcode'):
                self.linkcode = int(args.linkcode)
            if hasattr(args, 'wait'):
                self.wait_time = int(args.wait)

        if self.logfile:
            logging.basicConfig(level=self.loglevel, filename=self.logfile)
        else:
            _LOGGING.setLevel(self.loglevel)

    @asyncio.coroutine
    def connect(self, poll_devices=False, device=None, workdir=None):
        yield from self.aldb_load_lock.acquire()
        _LOGGING.info('Connecting to Insteon PLM at %s', self.device)
        self.device = device if device else self.device
        self.workdir = workdir if workdir else self.workdir
        conn = yield from insteonplm.Connection.create(
            device=self.device,
            host=self.host,
            port=self.port,
            username=self.username,
            password=self.password,
            loop=self.loop,
            poll_devices=poll_devices,
            workdir=self.workdir)
        conn.protocol.add_device_callback(self.async_new_device_callback)
        conn.protocol.add_all_link_done_callback(
            self.async_aldb_loaded_callback)
        self.plm = conn.protocol
        yield from self.aldb_load_lock
        if self.aldb_load_lock.locked():
            self.aldb_load_lock.release()

    @asyncio.coroutine
    def monitor_mode(self, poll_devices=False, device=None, workdir=None):
        print("Running monitor mode")
        yield from self.connect(poll_devices=False, device=None, workdir=None)
        self.plm.monitor_mode()

    def async_new_device_callback(self, device):
        """Log that our new device callback worked."""
        _LOGGING.info(
            'New Device: %s cat: 0x%02x subcat: 0x%02x desc: %s, model: %s',
            device.id, device.cat, device.subcat,
            device.description, device.model)
        for state in device.states:
            device.states[state].register_updates(
                self.async_state_change_callback)

    def async_state_change_callback(self, id, state, value):
        _LOGGING.info('Device %s state %s value is changed to %02x',
                      id, state, value)

    def async_aldb_loaded_callback(self):
        if self.aldb_load_lock.locked():
            self.aldb_load_lock.release()
        _LOGGING.info('ALDB Loaded')

    @asyncio.coroutine
    def start_all_linking(self, linkcode, group, address=None):
        _LOGGING.info('Starting the All-Linking process')
        if address:
            linkdevice = self.plm.devices[Address(address).id]
            if not linkdevice:
                linkdevice = Device.create(self.plm, address, None, None)
            _LOGGING.info('Attempting to link the PLM to device %s. ',
                          address)
            self.plm.start_all_linking(linkcode, group)
            asyncio.sleep(.5, loop=self.loop)
            linkdevice.enter_linking_mode(group=group)
        else:
            _LOGGING.info('Starting All-Linking on PLM. '
                          'Waiting for button press')
            self.plm.start_all_linking(linkcode, group)
            yield from asyncio.sleep(self.wait_time, loop=self.loop)

        _LOGGING.info('%d devices added to the All-Link Database',
                      len(self.plm.devices))
        yield from asyncio.sleep(.1, loop=self.loop)

    def list_devices(self):
        """List devices in the ALDB."""
        if self.plm.devices:
            for addr in self.plm.devices:
                device = self.plm.devices[addr]
                if device.address.is_x10:
                    _LOGGING.info('Device: %s %s', device.address.human,
                                  device.description)
                else:
                    _LOGGING.info('Device: %s cat: 0x%02x subcat: 0x%02x '
                                  'desc: %s, model: %s',
                                  device.address.human, device.cat,
                                  device.subcat, device.description,
                                  device.model)
        else:
            _LOGGING.info('No devices found')
            if not self.plm.transport:
                _LOGGING.info('IM connection has not been made.')
                _LOGGING.info('Use `connect [device]` to open the connection')

    @asyncio.coroutine
    def on_off_test(self, addr, group):
        """Test the on/off method of a device.

        Usage:
            on_off_test address [group]
        Arguments:
            address: Required - INSTEON address of the device
            group: Optional - All-Link group number. Defaults to 1
        """
        device = None
        state = None
        if addr:
            dev_addr = Address(addr)
            device = self.plm.devices[dev_addr.id]

        if device:
            state = device.states[group]

        if state:
            if hasattr(state, 'on') and hasattr(state, 'off'):
                _LOGGING.info('Send on request')
                _LOGGING.info('----------------------')
                device.states[group].on()
                yield from asyncio.sleep(2, loop=self.loop)

                _LOGGING.info('Send off request')
                _LOGGING.info('----------------------')
                device.states[group].off()
                yield from asyncio.sleep(2, loop=self.loop)

                _LOGGING.info('Send on request')
                _LOGGING.info('----------------------')
                device.states[group].on()
                yield from asyncio.sleep(2, loop=self.loop)

                _LOGGING.info('Send off request')
                _LOGGING.info('----------------------')
                device.states[group].off()
                yield from asyncio.sleep(2, loop=self.loop)
            else:
                _LOGGING.warn('Device %s with state %d is not an on/off'
                              'device.', device.id, state.name)

        else:
            _LOGGING.error('Could not find device %s', addr)

    def print_device_aldb(self, addr):
        """Diplay the All-Link database for a device."""
        if Address(addr).id == self.plm.address.id:
            device = self.plm
        else:
            dev_addr = Address(addr)
            device = self.plm.devices[dev_addr.id]
        if device:
            if device.aldb.status in [ALDBStatus.LOADED, ALDBStatus.PARTIAL]:
                if device.aldb.status == ALDBStatus.PARTIAL:
                    _LOGGING.info('ALDB partially loaded for device %s', addr)
                for mem_addr in device.aldb:
                    record = device.aldb[mem_addr]
                    _LOGGING.debug('mem_addr: %s', mem_addr)
                    _LOGGING.info('ALDB record: %s', record)
            else:
                _LOGGING.info('ALDB not loaded. '
                              'Use `load_aldb %s` first.',
                              device.address.id)
        else:
            _LOGGING.info('Device not found.')

    def print_all_aldb(self):
        """Diplay the All-Link database for all devices."""
        addr = self.plm.address.id
        _LOGGING.info('ALDB for PLM device %s', addr)
        self.print_device_aldb(addr)
        if self.plm.devices:
            for addr in self.plm.devices:
                _LOGGING.info('ALDB for device %s', addr)
                self.print_device_aldb(addr)
        else:
            _LOGGING.info('No devices found')
            if not self.plm.transport:
                _LOGGING.info('IM connection has not been made.')
                _LOGGING.info('Use `connect [device]` to open the connection')

    @asyncio.coroutine
    def load_device_aldb(self, addr, clear=True):
        """Read the device ALDB."""
        dev_addr = Address(addr)
        device = None
        if dev_addr == self.plm.address:
            device = self.plm
        else:
            device = self.plm.devices[dev_addr.id]
        if device:
            if clear:
                device.aldb.clear()
            device.read_aldb()
            yield from asyncio.sleep(1, loop=self.loop)
            while device.aldb.status == ALDBStatus.LOADING:
                yield from asyncio.sleep(1, loop=self.loop)
            if device.aldb.status == ALDBStatus.LOADED:
                _LOGGING.info('ALDB loaded for device %s', addr)
            self.print_device_aldb(addr)
        else:
            _LOGGING.error('Could not find device %s', addr)

    @asyncio.coroutine
    def load_all_aldb(self, clear=True):
        """Read all devices ALDB."""
        for addr in self.plm.devices:
            yield from self.load_device_aldb(addr, clear)

    @asyncio.coroutine
    def write_aldb(self, addr, mem_addr: int, mode: str, group: int, target,
                   data1=0x00, data2=0x00, data3=0x00):
        """Write a device All-Link record."""
        dev_addr = Address(addr)
        target_addr = Address(target)
        device = self.plm.devices[dev_addr.id]
        if device:
            _LOGGING.debug('calling device write_aldb')
            device.write_aldb(mem_addr, mode, group, target_addr,
                              data1, data2, data3)
            yield from asyncio.sleep(1, loop=self.loop)
            while device.aldb.status == ALDBStatus.LOADING:
                yield from asyncio.sleep(1, loop=self.loop)
            self.print_device_aldb(addr)

    @asyncio.coroutine
    def del_aldb(self, addr, mem_addr: int):
        """Write a device All-Link record."""
        dev_addr = Address(addr)
        device = self.plm.devices[dev_addr.id]
        if device:
            _LOGGING.debug('calling device del_aldb')
            device.del_aldb(mem_addr)
            yield from asyncio.sleep(1, loop=self.loop)
            while device.aldb.status == ALDBStatus.LOADING:
                yield from asyncio.sleep(1, loop=self.loop)
            self.print_device_aldb(addr)

    def add_device_override(self, addr, cat, subcat, firmware=None):
        """Add a device override to the PLM."""
        self.plm.devices.add_override(addr, 'cat', cat)
        self.plm.devices.add_override(addr, 'subcat', subcat)
        if firmware:
            self.plm.devices.add_override(addr, 'firmware', firmware)

    def add_x10_device(self, housecode, unitcode, dev_type):
        """Add an X10 device to the PLM."""
        device = None
        try:
            device = self.plm.devices.add_x10_device(self.plm, housecode,
                                                     unitcode, dev_type)
        except ValueError:
            pass
        return device


class Commander(object):
    """Command object to manage itneractive sessions."""

    def __init__(self, loop, args=None):
        self.loop = loop

        self.tools = Tools(loop, args)
        self.stdout = sys.stdout

    def start(self):
        self.loop.create_task(self._read_line())
        self.loop.create_task(self._greeting())

    @asyncio.coroutine
    def _read_line(self):
        while True:
            input = yield from self.loop.run_in_executor(None,
                                                         sys.stdin.readline)
            yield from self._exec_cmd(input)
            self.stdout.write(PROMPT)
            sys.stdout.flush()

    @asyncio.coroutine
    def _exec_cmd(self, input):
        return_val = None
        func = None
        if input.strip():
            command, arg = self._parse_cmd(input)
            if command is None:
                return self._invalid(input)
            if command == '':
                return self._invalid(input)
            else:
                try:
                    func = getattr(self, 'do_' + command)
                except AttributeError:
                    func = self._invalid
                    arg = str(input)
                except KeyboardInterrupt:
                    func = None  # func(arg)
            if func:
                if asyncio.iscoroutinefunction(func):
                    return_val = yield from func(arg)
                else:
                    return_val = func(arg)
        return return_val

    def _parse_cmd(self, input):
        input = input.strip()
        command = None
        arg = None
        if input:
            i, n = 0, len(input)
            while i < n and input[i] in ALLOWEDCHARS:
                i += 1
            command = input[:i]
            arg = input[i:].strip()
        return command, arg

    @staticmethod
    def _invalid(input):
        _LOGGING.error("Invalid command: %s", input[:-1])

    async def _greeting(self):
        _LOGGING.info(INTRO)
        self.stdout.write(PROMPT)
        self.stdout.flush()

    @asyncio.coroutine
    def do_connect(self, args):
        """Connect to the PLM device.

        Usage:
            connect [device [workdir]]
        Arguments:
            device: PLM device (default /dev/ttyUSB0)
            workdir: Working directory to save and load device information
        """
        params = args.split()
        device = '/dev/ttyUSB0'
        workdir = None

        try:
            device = params[0]
        except IndexError:
            if self.tools.device:
                device = self.tools.device
        try:
            workdir = params[1]
        except IndexError:
            if self.tools.workdir:
                workdir = self.tools.workdir

        if device:
            yield from self.tools.connect(False, device=device,
                                          workdir=workdir)
        _LOGGING.info('Connection complete.')

    def do_running_tasks(self, arg):
        """List tasks running in the background.

        Usage:
            running_tasks
        Arguments:
        """
        for task in asyncio.Task.all_tasks(loop=self.loop):
            _LOGGING.info(task)

    @asyncio.coroutine
    def do_on_off_test(self, args):
        """Test the on/off method of a device.

        Usage:
            on_off_test address [group]

        Arguments:
            address: Required - INSTEON address of the device
            group: Optional - All-Link group number. Defaults to 1
        """
        params = args.split()
        addr = None
        group = None

        try:
            addr = params[0]
        except IndexError:
            addr = None

        try:
            group = int(params[1])
        except ValueError:
            group = None
        except IndexError:
            group = 1

        if addr and group:
            yield from self.tools.on_off_test(addr, group)
        else:
            _LOGGING.error('Invalid address or group')
            self.do_help('on_off_test')

    def do_list_devices(self, args):
        """List devices loaded in the IM.

        Usage:
            list_devices
        Arguments:
        """
        self.tools.list_devices()

    def do_add_all_link(self, args):
        """Add an All-Link record to the IM and a device.

        Usage:
            add_all_link [linkcode] [group] [address]
        Arguments:
            linkcode: 0 - PLM is responder
                      1 - PLM is controller
                      3 - PLM is controller or responder
                      Default 1
            group: All-Link group number (0 - 255). Default 0.
            address: INSTEON device to link with (not supported by all devices)
        """
        linkcode = 1
        group = 0
        addr = None
        params = args.split()
        if params:
            try:
                linkcode = int(params[0])
            except IndexError:
                linkcode = 1
            except ValueError:
                linkcode = None

            try:
                group = int(params[1])
            except IndexError:
                group = 0
            except ValueError:
                group = None

            try:
                addr = params[2]
            except IndexError:
                addr = None

        if linkcode in [0, 1, 3] and group >= 0 and group <= 255:
            self.loop.create_task(
                self.tools.start_all_linking(linkcode, group, addr))
        else:
            _LOGGING.error('Link code %d or group number %d not valid',
                           linkcode, group)
            self.do_help('add_all_link')

    def do_del_all_link(self, args):
        """Delete an All-Link record to the IM and a device.

        Usage:
            add_all_link [group] [address]
        Arguments:
            group: All-Link group number (0 - 255). Default 0.
            address: INSTEON device to unlink (not supported by all devices)
        """
        linkcode = 255
        group = 0
        addr = None
        params = args.split()
        if params:
            try:
                group = int(params[0])
            except IndexError:
                group = 0
            except ValueError:
                group = None

            try:
                addr = params[1]
            except IndexError:
                addr = None

        if group and group >= 0 and group <= 255:
            self.loop.create_task(
                self.tools.start_all_linking(linkcode, group, addr))
        else:
            _LOGGING.error('Group number not valid')
            self.do_help('del_all_link')

    def do_print_aldb(self, args):
        """Print the All-Link database for a device.

        Usage:
            print_aldb address|plm|all
        Arguments:
            address: INSTEON address of the device
            plm: Print the All-Link database for the PLM
            all: Print the All-Link database for all devices

        This method requires that the device ALDB has been loaded.
        To load the device ALDB use the command:
            load_aldb address|plm|all
        """
        params = args.split()
        addr = None

        try:
            addr = params[0]
        except IndexError:
            _LOGGING.error('Device address required.')
            self.do_help('print_aldb')

        if addr:
            if addr.lower() == 'all':
                self.tools.print_all_aldb()
            elif addr.lower() == 'plm':
                addr = self.tools.plm.address.id
                self.tools.print_device_aldb(addr)
            else:
                self.tools.print_device_aldb(addr)

    def do_set_hub_connection(self, args):
        """Set Hub connection parameters.

        Usage:
            set_hub_connection username password host [port]

        Arguments:
            username: Hub username
            password: Hub password
            host: host name or IP address
            port: IP port [default 25105]
        """
        params = args.split()
        username = None
        password = None
        host = None
        port = None

        try:
            username = params[0]
            password = params[1]
            host = params[2]
            port = params[3]
        except IndexError:
            pass

        if username and password and host:
            if not port:
                port = 25105
            self.tools.username = username
            self.tools.password = password
            self.tools.host = host
            self.tools.port = port
        else:
            _LOGGING.error('username password host are required')
            self.do_help('set_hub_connection')

    def do_set_log_file(self, args):
        """Set the log file.

        Usage:
            set_log_file filename
        Parameters:
            filename: log file name to write to

        THIS CAN ONLY BE CALLED ONCE AND MUST BE CALLED
        BEFORE ANY LOGGING STARTS.
        """
        params = args.split()
        try:
            filename = params[0]
            logging.basicConfig(filename=filename)
        except IndexError:
            self.do_help('set_log_file')

    @asyncio.coroutine
    def do_load_aldb(self, args):
        """Load the All-Link database for a device.

        Usage:
            load_aldb address|all [clear_prior]
        Arguments:
            address: NSTEON address of the device
            all: Load the All-Link database for all devices
            clear_prior: y|n
                         y - Clear the prior data and start fresh.
                         n - Keep the prior data and only apply changes
                         Default is y
        This does NOT write to the database so no changes are made to the
        device with this command.
        """
        params = args.split()
        addr = None
        clear = True

        try:
            addr = params[0]
        except IndexError:
            _LOGGING.error('Device address required.')
            self.do_help('load_aldb')

        try:
            clear_prior = params[1]
            _LOGGING.info('param clear_prior %s', clear_prior)
            if clear_prior.lower() == 'y':
                clear = True
            elif clear_prior.lower() == 'n':
                clear = False
            else:
                _LOGGING.error('Invalid value for parameter `clear_prior`')
                _LOGGING.error('Valid values are `y` or `n`')
        except IndexError:
            pass

        if addr:
            if addr.lower() == 'all':
                yield from self.tools.load_all_aldb(clear)
            else:
                yield from self.tools.load_device_aldb(addr, clear)
        else:
            self.do_help('load_aldb')

    @asyncio.coroutine
    def do_write_aldb(self, args):
        """Write device All-Link record.

        WARNING THIS METHOD CAN DAMAGE YOUR DEVICE IF USED INCORRECTLY.
        Please ensure the memory id is appropriate for the device.
        You must load the ALDB of the device before using this method.
        The memory id must be an existing memory id in the ALDB or this
        method will return an error.

        If you are looking to create a new link between two devices,
        use the `link_devices` command or the `start_all_linking` command.

        Usage:
           write_aldb addr memory mode group target [data1 data2 data3]

        Required Parameters:
            addr: Inseon address of the device to write
            memory: record ID of the record to write (i.e. 0fff)
            mode: r | c
                    r = Device is a responder of target
                    c = Device is a controller of target
            group:  All-Link group integer
            target: Insteon address of the link target device

        Optional Parameters:
            data1: int = Device sepcific
            data2: int = Device specific
            data3: int = Device specific
        """
        params = args.split()
        addr = None
        mem_bytes = None
        memory = None
        mode = None
        group = None
        target = None
        data1 = 0x00
        data2 = 0x00
        data3 = 0x00

        try:
            addr = Address(params[0])
            mem_bytes = binascii.unhexlify(params[1])
            memory = int.from_bytes(mem_bytes, byteorder='big')
            mode = params[2]
            group = int(params[3])
            target = Address(params[4])

            _LOGGING.info('address: %s', addr)
            _LOGGING.info('memory: %04x', memory)
            _LOGGING.info('mode: %s', mode)
            _LOGGING.info('group: %d', group)
            _LOGGING.info('target: %s', target)

        except IndexError:
            _LOGGING.error('Device address memory mode group and target '
                           'are all required.')
            self.do_help('write_aldb')
        except ValueError:
            _LOGGING.error('Value error - Check parameters')
            self.do_help('write_aldb')

        try:
            data1 = int(params[5])
            data2 = int(params[6])
            data3 = int(params[7])
        except IndexError:
            pass
        except ValueError:
            addr = None
            _LOGGING.error('Value error - Check parameters')
            self.do_help('write_aldb')
            return

        if addr and memory and mode and isinstance(group, int) and target:
            yield from self.tools.write_aldb(addr, memory, mode, group, target,
                                             data1, data2, data3)

    @asyncio.coroutine
    def do_del_aldb(self, args):
        """Delete device All-Link record.

        WARNING THIS METHOD CAN DAMAGE YOUR DEVICE IF USED INCORRECTLY.
        Please ensure the memory id is appropriate for the device.
        You must load the ALDB of the device before using this method.
        The memory id must be an existing memory id in the ALDB or this
        method will return an error.

        If you are looking to create a new link between two devices,
        use the `link_devices` command or the `start_all_linking` command.

        Usage:
           del_aldb addr memory

        Required Parameters:
            addr: Inseon address of the device to write
            memory: record ID of the record to write (i.e. 0fff)
        """
        params = args.split()
        addr = None
        mem_bytes = None
        memory = None

        try:
            addr = Address(params[0])
            mem_bytes = binascii.unhexlify(params[1])
            memory = int.from_bytes(mem_bytes, byteorder='big')

            _LOGGING.info('address: %s', addr)
            _LOGGING.info('memory: %04x', memory)

        except IndexError:
            _LOGGING.error('Device address and memory are required.')
            self.do_help('del_aldb')
        except ValueError:
            _LOGGING.error('Value error - Check parameters')
            self.do_help('write_aldb')
        
        if addr and memory:
            yield from self.tools.del_aldb(addr, memory)


    def do_set_log_level(self, arg):
        """Set the log level.

        Usage:
            set_log_level i|v
        Parameters:
            log_level: i - info
                       v - verbose
    """
        if arg in ['i', 'v']:
            if arg == 'i':
                _LOGGING.setLevel(logging.INFO)
            else:
                _LOGGING.setLevel(logging.DEBUG)
        else:
            _LOGGING.error('Log level value error.')
            self.do_help('set_log_level')

    def do_set_device(self, args):
        """Set the PLM OS device.

        Device defaults to /dev/ttyUSB0

        Usage:
            set_device device
        Arguments:
            device: Required - INSTEON PLM device
        """
        params = args.split()
        device = None

        try:
            device = params[0]
        except IndexError:
            _LOGGING.error('Device name required.')
            self.do_help('set_device')

        if device:
            self.tools.device = device

    def do_set_workdir(self, args):
        """Set the working directory.

        The working directory is used to load and save known devices
        to improve startup times. During startup the application
        loads and saves a file `insteon_plm_device_info.dat`. This file
        is saved in the working directory.

        The working directory has no default value. If the working directory is
        not set, the `insteon_plm_device_info.dat` file is not loaded or saved.

        Usage:
            set_workdir workdir
        Arguments:
            workdir: Required - Working directory to load and save devie list
        """
        params = args.split()
        workdir = None

        try:
            workdir = params[0]
        except IndexError:
            _LOGGING.error('Device name required.')
            self.help('set_workdir')

        if workdir:
            self.tools.workdir = workdir

    def do_help(self, arg):
        """Help command.

        Usage:
            help [command]
        Parameters:
            command: Optional - command name to display detailed help
        """
        cmds = arg.split()

        if cmds:
            func = getattr(self, 'do_{}'.format(cmds[0]))
            if func:
                _LOGGING.info(func.__doc__)
            else:
                _LOGGING.error('Command ', cmds[0], ' not found.')
        else:
            _LOGGING.info("Available command list: ")
            for curr_cmd in dir(self.__class__):
                if curr_cmd.startswith("do_") and not curr_cmd == 'do_test':
                    print(" - ", curr_cmd[3:])
            _LOGGING.info("For help with a command type `help command`")

    def do_exit(self, arg):
        """Exit the application."""
        _LOGGING.info("Exiting")
        raise KeyboardInterrupt

    def do_test_logger(self, arg):
        _LOGGING.error("This is an error")
        _LOGGING.warn("This is a warn")
        _LOGGING.warning("This is a warning")
        _LOGGING.info("This is an info")
        _LOGGING.debug("This is a debug")

    def do_add_device_override(self, args):
        """Add a device override to the IM.

        Usage:
            add_device_override address cat subcat [firmware]

        Arguments:
            address: Insteon address of the device to override
            cat: Device category
            subcat: Device subcategory
            firmware: Optional - Device firmware

        The device address can be written with our without the dots and in
        upper or lower case, for example: 1a2b3c or 1A.2B.3C.

        The category, subcategory and firmware numbers are written in hex
        format, for example: 0x01 0x1b

        Example:
            add_device_override 1a2b3c 0x02 0x1a
        """
        params = args.split()
        addr = None
        cat = None
        subcat = None
        firmware = None
        error = None

        try:
            addr = Address(params[0])
            cat = binascii.unhexlify(params[1][2:])
            subcat = binascii.unhexlify(params[2][2:])
            firmware = binascii.unhexlify(params[3][2:])
        except IndexError:
            error = 'missing'
        except ValueError:
            error = 'value'

        if addr and cat and subcat:
            self.tools.add_device_override(addr, cat, subcat, firmware)
        else:
            if error == 'missing':
                _LOGGING.error('Device address, category and subcategory are '
                               'required.')
            else:
                _LOGGING.error('Check the vales for address, category and '
                               'subcategory.')
            self.do_help('add_device_override')

    def do_add_x10_device(self, args):
        """Add an X10 device to the IM.

        Usage:
            add_x10_device housecode unitcode type

        Arguments:
            housecode: Device housecode (A - P)
            unitcode: Device unitcode  (1 - 16)
            type: Device type

        Current device types are:
            - OnOff
            - Dimmable
            - Sensor

        Example:
            add_x10_device M 12 OnOff
        """
        params = args.split()
        housecode = None
        unitcode = None
        dev_type = None

        try:
            housecode = params[0]
            unitcode = int(params[1])
            if unitcode not in range(1, 17):
                raise ValueError
            dev_type = params[2]
        except IndexError:
            pass
        except ValueError:
            _LOGGING.error('X10 unit code must be an integer 1 - 16')
            unitcode = None

        if housecode and unitcode and dev_type:
            device = self.tools.add_x10_device(housecode, unitcode, dev_type)
            if not device:
                _LOGGING.error('Device not added. Please check the '
                               'information you provided.')
                self.do_help('add_x10_device')
        else:
            _LOGGING.error('Device housecode, unitcode and type are '
                           'required.')
            self.do_help('add_x10_device')


def monitor():
    """Connect to receiver and show events as they occur.

    Pulls the following arguments from the command line:

    :param device:
        Unix device where the PLM is attached
    :param address:
        Insteon address of the device to link with
    :param group:
        Insteon group for the link
    :param: linkcode
        Link direction: 0 - PLM is responder
                        1 - PLM is controller
                        3 - IM is responder or controller'
    :param verbose:
        Show debug logging.
    """
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('--device', default='/dev/ttyUSB0',
                        help='Path to PLM device')
    parser.add_argument('--verbose', '-v', action='count',
                        help='Set logging level to verbose')
    parser.add_argument('-l', '--logfile', default='',
                        help='Log file name')
    parser.add_argument('--workdir', default='',
                        help='Working directory for reading and saving '
                        'device information.')

    args = parser.parse_args()
    loop = asyncio.get_event_loop()
    monTool = Tools(loop, args)
    asyncio.ensure_future(monTool.monitor_mode())
    try:
        loop.run_forever()
    except KeyboardInterrupt:
        if cmd.tools.plm:
            if cmd.tools.plm.transport:
                _LOGGING.info('Closing the session')
                cmd.tools.plm.transport.close()
        loop.stop()
        pending = asyncio.Task.all_tasks(loop=loop)
        for task in pending:
            task.cancel()
            try:
                loop.run_until_complete(task)
            except asyncio.CancelledError:
                pass
            except KeyboardInterrupt:
                pass
        loop.close()


def interactive():
    """Wrapper for an interactive session for manual commands to be entered."""

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('--device', default='/dev/ttyUSB0',
                        help='Path to PLM device')
    parser.add_argument('-v', '--verbose', action='count',
                        help='Set logging level to verbose')
    parser.add_argument('-l', '--logfile', default='',
                        help='Log file name')
    parser.add_argument('--workdir', default='',
                        help='Working directory for reading and saving '
                        'device information.')
    args = parser.parse_args()

    loop = asyncio.get_event_loop()
    cmd = Commander(loop, args)
    cmd.start()
    try:
        loop.run_forever()
    except KeyboardInterrupt:
        if cmd.tools.plm:
            if cmd.tools.plm.transport:
                _LOGGING.info('Closing the session')
                cmd.tools.plm.transport.close()
        loop.stop()
        pending = asyncio.Task.all_tasks(loop=loop)
        for task in pending:
            task.cancel()
            try:
                loop.run_until_complete(task)
            except asyncio.CancelledError:
                pass
            except KeyboardInterrupt:
                pass
        loop.close()



"""Module to maintain PLM state information and network interface."""

import asyncio
import json
import logging

from insteonplm.address import Address
import insteonplm.devices
from insteonplm.devices import Device, X10Device

__all__ = ('ALDB')

DEVICE_INFO_FILE = 'insteon_plm_device_info.dat'


class LinkedDevices(object):
    """Class holds and maintains the ALL-Link Database from the PLM device."""

    def __init__(self, loop=None, workdir=None):
        """Instantiate the ALL-Link Database object."""
        self.log = logging.getLogger(__name__)
        self._loop = loop
        self._workdir = workdir

        self._state = 'empty'
        self._devices = {}
        self._cb_new_device = []
        self._overrides = {}
        self._saved_devices = {}

    def __len__(self):
        """Return the number of devices in the ALDB."""
        return len(self._devices)

    def __iter__(self):
        """Iterate through each ALDB device record."""
        for device in self._devices:
            yield device

    def __getitem__(self, address):
        """Fetch a device from the ALDB."""
        return self._devices.get(address, None)

    def __setitem__(self, key, device):
        """Add or Update a device in the ALDB."""
        if (not isinstance(device, Device) and
                not isinstance(device, X10Device)):
            raise ValueError

        self._devices[key] = device

        if device.address.is_x10:
            self.log.debug('New X10 Device %r: %s',
                           key, device.description)
        else:
            self.log.debug('New INSTEON Device %r: %s (%02x:%02x)',
                           key, device.description, device.cat,
                           device.subcat)

        for callback in self._cb_new_device:
            callback(device)

    def __repr__(self):
        """Human representation of a device from the ALDB."""
        attrs = vars(self)
        return ', '.join("%s: %r" % item for item in attrs.items())

    @property
    def saved_devices(self):
        """Return the device info from the saved devices file."""
        return self._saved_devices

    @property
    def overrides(self):
        """Return the device overrides."""
        return self._overrides

    @property
    def state(self):
        """Return the state of the ALDB.

        Possible states:
            empty
            loading
            loaded
        """
        return self._state

    @state.setter
    def state(self, val):
        """Set the ALDB load state."""
        self._state = val

    def add_device_callback(self, callback):
        """Register a callback to be invoked when a new device appears."""
        self.log.debug('Added new callback %s ',
                       callback)
        self._cb_new_device.append(callback)

    def add_override(self, addr, key, value):
        """Register an attribute override for a device."""
        address = Address(str(addr)).id
        self.log.info('New override for %s %s is %s', address, key, value)
        device_override = self._overrides.get(address, {})
        device_override[key] = value
        self._overrides[address] = device_override

    def create_device_from_category(self, plm, addr, cat, subcat,
                                    product_key=0x00):
        """Create a new device from the cat, subcat and product_key data."""
        saved_device = self._saved_devices.get(Address(addr).id, {})
        cat = saved_device.get('cat', cat)
        subcat = saved_device.get('subcat', subcat)
        product_key = saved_device.get('product_key', product_key)

        device_override = self._overrides.get(Address(addr).id, {})
        cat = device_override.get('cat', cat)
        subcat = device_override.get('subcat', subcat)
        product_key = device_override.get('firmware', product_key)
        product_key = device_override.get('product_key', product_key)

        return insteonplm.devices.create(plm, addr, cat, subcat, product_key)

    def has_saved(self, addr):
        """Test if device has data from the saved data file."""
        saved = False
        if self._saved_devices.get(addr, None) is not None:
            saved = True
        return saved

    def has_override(self, addr):
        """Test if device has data from a device override setting."""
        override = False
        if self._overrides.get(addr, None) is not None:
            override = True
        return override

    def add_known_devices(self, plm):
        """Add devices from the saved devices or from the device overrides."""
        from insteonplm.devices import ALDBStatus
        for addr in self._saved_devices:
            if not self._devices.get(addr):
                saved_device = self._saved_devices.get(Address(addr).id, {})
                cat = saved_device.get('cat')
                subcat = saved_device.get('subcat')
                product_key = saved_device.get('firmware')
                product_key = saved_device.get('product_key', product_key)
                device = self.create_device_from_category(
                        plm, addr, cat, subcat, product_key)
                if device:
                    self.log.info('Device with id %s added to device list '
                                  'from saved device data.',
                                  addr)
                    aldb_status = saved_device.get('aldb_status', 0)
                    device.aldb.status = ALDBStatus(aldb_status)
                    aldb = saved_device.get('aldb', {})
                    device.aldb.load_saved_records(aldb_status, aldb)
                    self[addr] = device
        for addr in self._overrides:
            if not self._devices.get(addr):
                device_override = self._overrides.get(Address(addr).id, {})
                cat = device_override.get('cat')
                subcat = device_override.get('subcat')
                product_key = device_override.get('firmware')
                product_key = device_override.get('product_key', product_key)
                device = self.create_device_from_category(
                        plm, addr, cat, subcat, product_key)
                if device:
                    self.log.info('Device with id %s added to device list '
                                  'from device override data.',
                                  addr)
                    self[addr] = device

    # Save device information
    def save_device_info(self):
        if self._workdir is not None:
            devices = []
            for addr in self._devices:
                device = self._devices.get(addr)
                if not device.address.is_x10:
                    aldb = {}
                    for mem in device.aldb:
                        rec = device.aldb[mem]
                        if rec:
                            aldbRec = {'memory': mem,
                                       'control_flags': rec.control_flags.byte,
                                       'group': rec.group,
                                       'address': rec.address.id,
                                       'data1': rec.data1,
                                       'data2': rec.data2,
                                       'data3': rec.data3}
                            aldb[mem] = aldbRec
                    deviceInfo = {'address': device.address.id,
                                  'cat': device.cat,
                                  'subcat': device.subcat,
                                  'product_key': device.product_key,
                                  'aldb_status': device.aldb.status.value,
                                  'aldb': aldb}
                    devices.append(deviceInfo)
            coro = self._write_saved_device_info(devices)
            asyncio.ensure_future(coro, loop=self._loop)

    def _add_saved_device_info(self, **kwarg):
        """Register device info from the saved data file."""
        addr = kwarg.get('address')
        self.log.debug('Found saved device with address %s', addr)
        self._saved_devices[addr] = kwarg

    @asyncio.coroutine
    def load_saved_device_info(self):
        self.log.debug("Loading saved device info.")
        deviceinfo = []
        if self._workdir:
            self.log.debug("Really Loading saved device info.")
            try:
                device_file = '{}/{}'.format(self._workdir, DEVICE_INFO_FILE)
                with open(device_file, 'r') as infile:
                    try:
                        deviceinfo = json.load(infile)
                        self.log.debug("Saved device file loaded")
                    except json.decoder.JSONDecodeError:
                        self.log.debug("Loading saved device file failed")
            except FileNotFoundError:
                self.log.debug("Saved device file not found")
        for device in deviceinfo:
            self._add_saved_device_info(**device)

    @asyncio.coroutine
    def _write_saved_device_info(self, devices):
        if self._workdir:
            self.log.debug('Writing %d devices to save file', len(devices))
            device_file = '{}/{}'.format(self._workdir, DEVICE_INFO_FILE)
            try:
                with open(device_file, 'w') as outfile:
                    json.dump(devices, outfile)
            except FileNotFoundError:
                self.log.error('Cannot write to file %s', device_file)



"""Mock Connection for the PLM."""

import asyncio
import binascii
import logging


_LOGGER = logging.getLogger()


# pylint: disable=too-few-public-methods
class MockConnection():
    """A mock up of the Connection class."""

    def __init__(self):
        """Instantiate the Connection object."""
        self.loop = None
        self.protocol = None
        self.transport = None

    @classmethod
    @asyncio.coroutine
    def create(cls, loop=None):
        from insteonplm.plm import PLM
        """Create a mock connection."""
        conn = cls()
        conn.loop = loop or asyncio.get_event_loop()
        conn.protocol = PLM(
            connection_lost_callback=None,
            loop=conn.loop)

        # pylint: disable=too-few-public-methods
        class Serial:
            """Mock serial class within Connection class."""

            def __init__(self):
                """Initialize the mock Serial class."""
                self.write_timeout = 0
                self.timeout = 0

        class Transport:
            """Mock transport class within Connection class."""

            def __init__(self):
                """Initialize the mock Transport class."""
                self.serial = Serial()
                self.lastmessage = None
                self._mock_buffer_size = 128

            def set_write_buffer_limits(self, num):
                """Mock set write buffer limits."""
                pass

            def get_write_buffer_size(self):
                """Mock get write buffer size."""
                return self._mock_buffer_size

            def write(self, data):
                """Mock write data to the Connection."""
                self.lastmessage = binascii.hexlify(data).decode()
                _LOGGER.info('Message sent: %s', self.lastmessage)

        conn.transport = Transport()
        return conn


"""INSTEON Set IM Configuration Message."""

from insteonplm.messages.message import Message
from insteonplm.constants import (MESSAGE_SET_IM_CONFIGURATION_0X6B,
                                  MESSAGE_SET_IM_CONFIGURATION_SIZE,
                                  MESSAGE_SET_IM_CONFIGURATION_RECEIVED_SIZE,
                                  MESSAGE_ACK,
                                  MESSAGE_NAK)
from insteonplm.address import Address


class SetIMConfiguration(Message):
    """INSTEON Get Insteon Modem Info Message.

    Message type 0x60
    """

    _code = MESSAGE_SET_IM_CONFIGURATION_0X6B
    _sendSize = MESSAGE_SET_IM_CONFIGURATION_SIZE
    _receivedSize = MESSAGE_SET_IM_CONFIGURATION_RECEIVED_SIZE
    _description = 'INSTEON Set IM Configuration Message'

    def __init__(self, flags=None, acknak=None):
        """Initialize the GetImInfo Class."""
        self._imConfigurationFlags = flags
        self._acknak = self._setacknak(acknak)

    @classmethod
    def from_raw_message(cls, rawmessage):
        """Create message from raw byte stream."""
        return SetIMConfiguration(rawmessage[2],
                                  rawmessage[3])

    @property
    def imConfigurationFlags(self):
        """Return the IM configuration flags."""
        return self._imConfigurationFlags

    @property
    def acknak(self):
        """Return the ACK/NAK byte."""
        return self._acknak

    @property
    def isack(self):
        """Test if this is an ACK message."""
        return self._acknak is not None and self._acknak == MESSAGE_ACK

    @property
    def isnak(self):
        """Test if this is a NAK message."""
        return self._acknak is not None and self._acknak == MESSAGE_NAK

    def _message_properties(self):
        return [{'imConfigurationFlags': self._imConfigurationFlags},
                {'acknak': self.acknak}]




"""X10 states."""
import logging

from insteonplm.messages.x10send import X10Send
from insteonplm.messages.x10received import X10Received
from insteonplm.states import State
from insteonplm.constants import (X10_COMMAND_ALL_UNITS_OFF,
                                  X10_COMMAND_ALL_LIGHTS_ON,
                                  X10_COMMAND_ALL_LIGHTS_OFF,
                                  X10_COMMAND_ON,
                                  X10_COMMAND_OFF,
                                  X10_COMMAND_DIM,
                                  X10_COMMAND_BRIGHT)


_LOGGER = logging.getLogger()


class X10OnOffSwitch(State):
    """On / Off state for an X10 device."""

    def __init__(self, address, statename, group, send_message_method,
                 message_callbacks, defaultvalue=None):
        """Initialize the X10OnOff state."""
        super().__init__(address, statename, group, send_message_method,
                         message_callbacks, defaultvalue)

        self._register_messages()

    def on(self):
        """Send the On command to an X10 device."""
        msg = X10Send.unit_code_msg(self.address.x10_housecode,
                                    self.address.x10_unitcode)
        self._send_method(msg)

        msg = X10Send.command_msg(self.address.x10_housecode,
                                  X10_COMMAND_ON)
        self._send_method(msg, False)
        self._update_subscribers(0xff)

    def off(self):
        """Send the Off command to an X10 device."""
        msg = X10Send.unit_code_msg(self.address.x10_housecode,
                                    self.address.x10_unitcode)
        self._send_method(msg)

        msg = X10Send.command_msg(self.address.x10_housecode,
                                  X10_COMMAND_OFF)
        self._send_method(msg, False)
        self._update_subscribers(0x00)

    def _on_message_received(self, msg):
        """An ON has been received."""
        self._update_subscribers(0xff)

    def _off_message_received(self, msg):
        """An OFF has been received."""
        self._update_subscribers(0x00)

    def _register_messages(self):
        on_msg = X10Received.command_msg(self.address.x10_housecode,
                                         X10_COMMAND_ON)
        off_msg = X10Received.command_msg(self.address.x10_housecode,
                                          X10_COMMAND_OFF)
        all_on_msg = X10Received.command_msg(self.address.x10_housecode,
                                             X10_COMMAND_ALL_LIGHTS_ON)
        all_off_msg = X10Received.command_msg(self.address.x10_housecode,
                                              X10_COMMAND_ALL_LIGHTS_OFF)
        all_units_off_msg = X10Received.command_msg(self.address.x10_housecode,
                                                    X10_COMMAND_ALL_UNITS_OFF)

        self._message_callbacks.add(on_msg,
                                    self._on_message_received)
        self._message_callbacks.add(off_msg,
                                    self._off_message_received)
        self._message_callbacks.add(all_on_msg,
                                    self._on_message_received)
        self._message_callbacks.add(all_off_msg,
                                    self._off_message_received)
        self._message_callbacks.add(all_units_off_msg,
                                    self._off_message_received)


class X10DimmableSwitch(X10OnOffSwitch):
    """Dimmable X10 Switch."""

    def __init__(self, address, statename, group, send_message_method,
                 message_callbacks, defaultvalue=0):
        """Initialize the Dimmable state."""
        super().__init__(address, statename, group, send_message_method,
                         message_callbacks, defaultvalue)

        self._steps = 22

    @property
    def steps(self):
        """Number of steps from OFF to full ON."""
        return self._div_steps

    @steps.setter
    def steps(self, val: int):
        """Set the number of steps from OFF to full ON."""
        self._dim_steps = val

    def set_level(self, val):
        """Set the device ON LEVEL."""
        if val == 0:
            self.off()
        elif val == 255:
            self.on()
        else:
            setlevel = 255
            if val < 1:
                setlevel = val*255
            elif val <= 0xff:
                setlevel = val
            change = setlevel - self._value
            increment = 255 / self._steps
            steps = round(abs(change) / increment)
            print('Steps: ', steps)
            if change > 0:
                method = self.brighten
                self._value += round(steps * increment)
                self._value = min(255, self._value)
            else:
                method = self.dim
                self._value -= round(steps * increment)
                self._value == max(0, self._value)
            for step in range(0, steps):
                method(True)
            self._update_subscribers(self._value)

    def brighten(self, defer_update=False):
        """Brighten the device one step."""
        msg = X10Send.unit_code_msg(self.address.x10_housecode,
                                    self.address.x10_unitcode)
        self._send_method(msg)

        msg = X10Send.command_msg(self.address.x10_housecode,
                                  X10_COMMAND_BRIGHT)
        self._send_method(msg, False)
        if not defer_update:
            self._update_subscribers(self._value + 255 / self._steps)

    def dim(self, defer_update=False):
        """Dim the device one step."""
        msg = X10Send.unit_code_msg(self.address.x10_housecode,
                                    self.address.x10_unitcode)
        self._send_method(msg)

        msg = X10Send.command_msg(self.address.x10_housecode,
                                  X10_COMMAND_DIM)
        self._send_method(msg, False)
        if not defer_update:
            self._update_subscribers(self._value - 255 / self._steps)

    def _dim_message_received(self, msg):
        val = max(self._value - (255 / self._steps), 0)
        self._update_subscribers(val)

    def _bright_message_received(self, msg):
        val = min(self._value + (255 / self._steps), 255)
        self._update_subscribers(val)

    def _register_messages(self):
        super()._register_messages()
        dim_msg = X10Received.command_msg(self.address.x10_housecode,
                                          X10_COMMAND_DIM)
        bri_msg = X10Received.command_msg(self.address.x10_housecode,
                                          X10_COMMAND_BRIGHT)
        self._message_callbacks.add(dim_msg,
                                    self._dim_message_received)
        self._message_callbacks.add(bri_msg,
                                    self._bright_message_received)


class X10OnOffSensor(State):
    """On / Off state for an X10 device."""

    def __init__(self, address, statename, group, send_message_method,
                 message_callbacks, defaultvalue=None):
        """Initialize the X10OnOff state."""
        super().__init__(address, statename, group, send_message_method,
                         message_callbacks, defaultvalue)

        self._register_messages()

    def _on_message_received(self, msg):
        """An ON has been received."""
        self._update_subscribers(0xff)

    def _off_message_received(self, msg):
        """An OFF has been received."""
        self._update_subscribers(0x00)

    def _register_messages(self):
        on_msg = X10Received.command_msg(self.address.x10_housecode,
                                         X10_COMMAND_ON)
        off_msg = X10Received.command_msg(self.address.x10_housecode,
                                          X10_COMMAND_OFF)
        all_units_off_msg = X10Received.command_msg(self.address.x10_housecode,
                                                    X10_COMMAND_ALL_UNITS_OFF)

        self._message_callbacks.add(on_msg,
                                    self._on_message_received)
        self._message_callbacks.add(off_msg,
                                    self._off_message_received)
        self._message_callbacks.add(all_units_off_msg,
                                    self._off_message_received)


class X10AllUnitsOffSensor(State):
    """All Units Off state for an X10 device."""

    def __init__(self, address, statename, group, send_message_method,
                 message_callbacks, defaultvalue=0xff):
        """Initialize the X10AllUnitsOff state."""
        super().__init__(address, statename, group, send_message_method,
                         message_callbacks, defaultvalue)

        self._register_messages()

    def reset(self):
        """Reset the state to ON"""
        self._update_subscribers(0xff)

    def _off_message_received(self, msg):
        """An OFF has been received."""
        self._update_subscribers(0x00)

    def _register_messages(self):
        all_units_off_msg = X10Received.command_msg(self.address.x10_housecode,
                                                    X10_COMMAND_ALL_UNITS_OFF)
        self._message_callbacks.add(all_units_off_msg,
                                    self._off_message_received)


class X10AllLightsOnSensor(State):
    """All Units Off state for an X10 device."""

    def __init__(self, address, statename, group, send_message_method,
                 message_callbacks, defaultvalue=0x00):
        """Initialize the X10AllLightsOn state."""
        super().__init__(address, statename, group, send_message_method,
                         message_callbacks, defaultvalue)

        self._register_messages()

    def reset(self):
        """Reset the state to OFF"""
        self._update_subscribers(0x00)

    def _on_message_received(self, msg):
        """An ON has been received."""
        self._update_subscribers(0xff)

    def _register_messages(self):
        all_on_msg = X10Received.command_msg(self.address.x10_housecode,
                                             X10_COMMAND_ALL_LIGHTS_ON)
        self._message_callbacks.add(all_on_msg,
                                    self._on_message_received)


class X10AllLightsOffSensor(State):
    """All Lights Off state for an X10 device."""

    def __init__(self, address, statename, group, send_message_method,
                 message_callbacks, defaultvalue=0xff):
        """Initialize the X10AllLightsOff state."""
        super().__init__(address, statename, group, send_message_method,
                         message_callbacks, defaultvalue)

        self._register_messages()

    def reset(self):
        """Reset the state to ON"""
        self._update_subscribers(0xff)

    def _off_message_received(self, msg):
        """An OFF has been received."""
        self._update_subscribers(0x00)

    def _register_messages(self):
        all_off_msg = X10Received.command_msg(self.address.x10_housecode,
                                              X10_COMMAND_ALL_LIGHTS_OFF)
        self._message_callbacks.add(all_off_msg,
                                    self._off_message_received)



"""INSTEON Message Class Module."""
import logging
import binascii
from insteonplm.address import Address
from insteonplm.constants import MESSAGE_START_CODE_0X02
from insteonplm.messages.messageFlags import MessageFlags
from insteonplm.messages.userdata import Userdata


class ClassPropertyMetaClass(type):
    """Meta class for Message class.

    This is meta class magic to allow class attributes to also appear as an
    instance property.
    """

    @property
    def code(cls):
        """Message code used to determine message type."""
        return cls._code

    @property
    def sendSize(cls):
        """Size of the message sent in bytes."""
        return cls._sendSize

    @property
    def receivedSize(cls):
        """Size of the message received in bytes."""
        return cls._receivedSize

    @property
    def description(cls):
        """Description of the message type."""
        return cls._description


class Message(metaclass=ClassPropertyMetaClass):
    """Base message class for an INSTEON message."""

    _code = 0
    _sendSize = 0
    _receivedSize = 0
    _description = "Empty message"
    log = logging.getLogger(__name__)

    def __str__(self):
        """String representation of an INSTEON message."""
        props = self._message_properties()
        msgstr = "{}'code': 0x{}".format(
            "{", binascii.hexlify(bytes([self._code])).decode())
        for prop in props:
            for key, val in prop.items():
                msgstr = msgstr + ', '
                if isinstance(val, Address):
                    msgstr = "{}'{}': {}".format(msgstr, key, val.human)
                elif isinstance(val, MessageFlags):
                    msgstr = "{}'{}': 0x{}".format(msgstr, key, val.hex)
                elif isinstance(val, int):
                    msgstr = "{}'{}': 0x{}".format(
                        msgstr, key, binascii.hexlify(bytes([val])).decode())
                elif isinstance(val, bytearray):
                    msgstr = "{}'{}': 0x{}".format(
                        msgstr, key, binascii.hexlify(val).decode())
                elif isinstance(val, bytes):
                    msgstr = "{}'{}': 0x{}".format(
                        msgstr, key, binascii.hexlify(val).decode())
                else:
                    msgstr = "{}'{}': 0x{}".format(msgstr, key, str(val))
        msgstr = "{}{}".format(msgstr, '}')
        return msgstr

    def __eq__(self, other):
        """Test for equality."""
        match = False
        if isinstance(other, Message) and other.code == self._code:
            match = str(self) == str(other)
        return match

    def __ne__(self, other):
        """Test for inequality."""
        match = True
        if isinstance(other, Message) and other.code == self._code:
            match = str(self) != str(other)
        return match

    def __lt__(self, other):
        """Test for less than."""
        less_than = False
        if isinstance(other, Message):
            less_than = str(self) < str(other)
        else:
            raise TypeError
        return less_than

    def __gt__(self, other):
        """Test for greater than."""
        greater_than = False
        if isinstance(other, Message):
            greater_than = str(self) > str(other)
        else:
            raise TypeError
        return greater_than

    def __hash__(self):
        """Create a has of the message."""
        return hash(str(self))

    @property
    def code(self):
        """Messasge code used to determine message type."""
        return self._code

    @property
    def sendSize(self):
        """Size of the sent messsage in bytes."""
        return self._sendSize

    @property
    def receivedSize(self):
        """Size of the received message in bytes."""
        return self._receivedSize

    @property
    def description(self):
        """Description of the message type."""
        return self._description

    @property
    def hex(self):
        """Hexideciaml representation of the message in bytes."""
        props = self._message_properties()
        msg = bytearray([MESSAGE_START_CODE_0X02, self._code])

        for prop in props:
            # pylint: disable=unused-variable
            for key, val in prop.items():
                if val is None:
                    pass
                elif isinstance(val, int):
                    msg.append(val)
                elif isinstance(val, Address):
                    if val.addr is None:
                        pass
                    else:
                        msg.extend(val.bytes)
                elif isinstance(val, MessageFlags):
                    msg.extend(val.bytes)
                elif isinstance(val, bytearray):
                    msg.extend(val)
                elif isinstance(val, bytes):
                    msg.extend(val)
                elif isinstance(val, Userdata):
                    msg.extend(val.bytes)

        return binascii.hexlify(msg).decode()

    @property
    def bytes(self):
        """Return the bytes representation of the message."""
        return binascii.unhexlify(self.hex)

    def matches_pattern(self, other):
        """Return if the current message matches a message template.

        Compare the current message to a template message to test matches
        to a pattern.
        """
        properties = self._message_properties()
        ismatch = False
        if isinstance(other, Message) and self.code == other.code:
            for prop in properties:
                for key, prop_val in prop.items():
                    if hasattr(other, key):
                        key_val = getattr(other, key)
                        ismatch = self._test_match(prop_val, key_val)
                    else:
                        ismatch = False
                    if not ismatch:
                        break
                if not ismatch:
                    break
        return ismatch

    @staticmethod
    def _setacknak(acknak):
        _acknak = acknak
        if isinstance(acknak, bytearray) and acknak:
            _acknak = acknak[0]
        return _acknak

    def _message_properties(self):
        raise NotImplementedError

    @staticmethod
    def _test_match(prop_val, key_val):
        ismatch = False
        if isinstance(prop_val, MessageFlags):
            ismatch = prop_val.matches_pattern(key_val)
        elif isinstance(prop_val, Address):
            ismatch = prop_val.matches_pattern(key_val)
        elif isinstance(prop_val, Userdata):
            ismatch = prop_val.matches_pattern(key_val)
        else:
            if prop_val is None or key_val is None:
                ismatch = True
            else:
                ismatch = prop_val == key_val
        return ismatch
